local VERSION = tonumber(("$Revision: 35239 $"):match("%d+"))

local PitBull = PitBull
if PitBull.revision < VERSION then
	PitBull.version = "r" .. VERSION
	PitBull.revision = VERSION
	PitBull.date = ("$Date: 2007-05-12 16:22:09 -0400 (Sat, 12 May 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local L = PitBull:L("PitBull-ABF")

local PitBull_Aura = nil

local DogTag = AceLibrary("DogTag-1.0")

local newFrame = PitBull.newFrame

local ABF = {}
PitBull.ABF = ABF

local ABF_db = PitBull:AcquireDBNamespace("Layout-ABF")
ABF.db = ABF
PitBull:RegisterDefaults("Layout-ABF", "profile", {
	["**"] = {
		portraitSide = "left",
		auraSide = "bottom",
		--note, these are only used when debuffs are split
		buffSide = "bottom",
		debuffSide = "right",
		
		classTextCustom = "[Classification] [Level:DifficultyColor] [SmartClass:ClassColor] [DruidForm:Paren] [SmartRace]",
		classTextStyle = "Standard",
		healthTextCustom = "[CurHP]/[MaxHP]",
		healthTextStyle = "Smart",
		nameTextCustom = "[Name]",
		nameTextStyle = "Standard",
		powerTextCustom = "[CurMP]/[MaxMP]",
		powerTextStyle = "Absolute",
		comboTextStyle = "Standard",
		experienceTextStyle = "Standard",
		repTextStyle = "Standard",
		
		healthBarHeight = 5,
		powerBarHeight = 4,
		castBarHeight = 2,
		repBarHeight = 2,
		expBarHeight = 2,
		druidManaBarHeight = 0.25,
		portraitHeight = 2,
	},
	player = {
		healthTextStyle = "Informational",
	},
	pet = {
		healthTextStyle = "Informational",
	},
	target = {
		portraitSide = "right",
		auraSide = "right",
		comboSide = "bottom",
		healthTextStyle = "Informational",
	}
})

local heights = {
	healthBar = 'healthBarHeight',
	powerBar = 'powerBarHeight',
	castBar = 'castBarHeight',
	repBar = 'repBarHeight',
	expBar = 'expBarHeight',
	portrait = 'portraitHeight',
}
ABF.heights = heights

local function figureHeight(multiplier, frame)
	local num = 0
	local db = ABF_db.profile[frame.group]
	for k,v in pairs(heights) do
		if frame[k] and frame[k]:IsShown() then
			if k == "portrait" then
				local portraitSide = ABF_db.profile[frame.group].portraitSide
				if portraitSide == "top" then
					num = num + db[v]
				end
			else
				num = num + db[v]
			end
		end
	end
	return db[heights[multiplier]] * frame:GetHeight() / num
end
ABF.figureHeight = figureHeight

function ABF.healthBar(unit, frame)
	local height = figureHeight('healthBar', frame)
	if frame.portrait then
		local portraitSide = ABF_db.profile[frame.group].portraitSide
		if portraitSide == "right" then
			frame.healthBar:SetPoint("TOP", frame, "TOP", 0, 0)
			frame.healthBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
			frame.healthBar:SetPoint("RIGHT", frame.portrait, "LEFT", -2, 0)
		elseif portraitSide == "left" then
			frame.healthBar:SetPoint("TOP", frame, "TOP", 0, 0)
			frame.healthBar:SetPoint("LEFT", frame.portrait, "RIGHT", 2, 0)
			frame.healthBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
		else -- top
			frame.healthBar:SetPoint("TOP", frame.portrait, "BOTTOM", 0, 0)
			frame.healthBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
			frame.healthBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
		end
	else
		frame.healthBar:SetPoint("TOP", frame, "TOP", 0, 0)
		frame.healthBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
		frame.healthBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
	end
	frame.healthBar:SetHeight(height) -- no pixel spacing, there is always 1 spacing less than bars.
end
function ABF.powerBar(unit, frame)
	local height = figureHeight('powerBar', frame)
	if frame.healthBar then
		frame.powerBar:SetPoint("TOP", frame.healthBar, "BOTTOM", 0, -1)
	else
		frame.powerBar:SetPoint("TOP", frame, "TOP", 0, 0)
	end
	if frame.portrait then
		local portraitSide = ABF_db.profile[frame.group].portraitSide
		if portraitSide == "right" then
			frame.powerBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
			frame.powerBar:SetPoint("RIGHT", frame.portrait, "LEFT", -2, 0)
		elseif portraitSide == "left" then
			frame.powerBar:SetPoint("LEFT", frame.portrait, "RIGHT", 2, 0)
			frame.powerBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
		else -- top
			frame.powerBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
			frame.powerBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
		end
	else
		frame.powerBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
		frame.powerBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
	end
	if frame.druidManaBar then
		height = height * (1 - ABF_db.profile[frame.group].druidManaBarHeight)
	end
	frame.powerBar:SetHeight(height - 1)
end
function ABF.druidManaBar(unit, frame)
	if not frame.powerBar then
		return
	end
	local height = figureHeight('powerBar', frame) * ABF_db.profile[frame.group].druidManaBarHeight
	frame.druidManaBar:SetPoint("TOPLEFT", frame.powerBar, "BOTTOMLEFT")
	frame.druidManaBar:SetPoint("TOPRIGHT", frame.powerBar, "TOPRIGHT")
	frame.druidManaBar:SetHeight(height)
end
function ABF.castBar(unit, frame)
	local height = figureHeight('castBar', frame)
	frame.castBar:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", height, 0)
	frame.castBar:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, 0)
	frame.castBar.bg:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 0, 0)
	frame.castBar.bg:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, 0)
	frame.castBar:SetHeight( height - 1) -- pixel spacing
	frame.castBar.bg:SetHeight(height - 1) -- pixel spacing
end
function ABF.repBar(unit, frame)
	local height = figureHeight('repBar', frame)
	frame.repBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
	frame.repBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
	if frame.expBar then
		frame.repBar:SetPoint("BOTTOM", frame.expBar, "TOP", 0, 1) -- 1 up spacing
	elseif frame.castBar then
		frame.repBar:SetPoint("BOTTOM", frame.castBar, "TOP", 0, 1) -- 1 up spacing
	else	
		frame.repBar:SetPoint("BOTTOM", frame, "BOTTOM", 0, 0)
	end	
	frame.repBar:SetHeight(height - 1)
end
function ABF.expBar(unit, frame)
	local height = figureHeight('expBar', frame)
	frame.expBar:SetPoint("LEFT", frame, "LEFT", 0, 0)
	frame.expBar:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
	if frame.castBar then
		frame.expBar:SetPoint("BOTTOM", frame.castBar, "TOP", 0, 1) -- 1 up spacing
	else
		frame.expBar:SetPoint("BOTTOM", frame, "BOTTOM", 0, 0)
	end	
	frame.expBar:SetHeight(height - 1)
end
function ABF.castBarIcon(unit, frame)
	local height = figureHeight('castBar', frame)
	frame.castBarIcon:SetHeight( height -1)
	frame.castBarIcon:SetWidth( height -1)
	frame.castBarIcon:SetPoint("RIGHT", frame.castBar, "LEFT" )
	frame.castBarIcon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
end
function ABF.castBarText(unit, frame)
	frame.castBarText:ClearAllPoints()
	frame.castBarText:SetAllPoints(frame.castBar)
	frame.castBarText:SetJustifyH("LEFT")
end
function ABF.castBarTimeText(unit, frame)
	frame.castBarTimeText:ClearAllPoints()
	frame.castBarTimeText:SetAllPoints(frame.castBar)
	frame.castBarTimeText:SetJustifyH("RIGHT")
end
function ABF.repBarText(unit, frame)
	frame.repBarText:ClearAllPoints()
	frame.repBarText:SetAllPoints(frame.repBar)
	frame.repBarText:SetJustifyH("CENTER")
end
function ABF.healthText(unit, frame)
	if frame.healthBar then
		frame.healthText:SetJustifyV("MIDDLE")
		frame.healthText:SetJustifyH("RIGHT")
		frame.healthText:SetPoint("RIGHT", frame.healthBar, "RIGHT", -3, 0)
	end
end
function ABF.powerText(unit, frame)
	if frame.powerBar then
		frame.powerText:SetJustifyV("MIDDLE")
		frame.powerText:SetJustifyH("RIGHT")
		frame.powerText:SetPoint("RIGHT", frame.powerBar, "RIGHT", -3, 0)
	end	
end
function ABF.nameText(unit, frame)
	if frame.healthBar then
		frame.nameText:SetPoint("LEFT", frame.healthBar, "LEFT", 3, 0)
		frame.nameText:SetJustifyV("MIDDLE")
		frame.nameText:SetJustifyH("LEFT")
		if frame.healthText then
			frame.nameText:SetPoint("RIGHT", frame.healthText, "LEFT", 0, 0)
		end
	end
end
function ABF.classText(unit, frame)
	if frame.powerBar then
		frame.classText:SetPoint("LEFT", frame.powerBar, "LEFT", 3, 0)
		frame.classText:SetJustifyV("MIDDLE")
		frame.classText:SetJustifyH("LEFT")
		if frame.powerText then
			frame.classText:SetPoint("RIGHT", frame.powerText, "LEFT", 0, 0)
		end
	end
end
function ABF.auraFrame(unit, frame)
	frame.auraFrame:ClearAllPoints()
	local size = PitBull:GetModule("Aura").db.profile[frame.group].buffsize
	frame.auraFrame:SetHeight(size)
	if PitBull:GetModule("Aura").db.profile[frame.group].split then
		local buffside = ABF_db.profile[frame.group].buffSide
		local w = frame:GetWidth()
		frame.auraFrame:SetWidth(w)
		if buffside == "bottom" then
			frame.auraFrame:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", -2, -2)
		elseif buffside == "top" then
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", -2, 2)
		elseif buffside == "left" then
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "LEFT", w-size-2, -2)
		elseif buffside == "right" then
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "RIGHT", 2, 0)
		elseif buffside == "topleft" then
			frame.auraFrame:SetPoint("TOPRIGHT", frame, "TOPLEFT", w-size-2, 0)
		elseif buffside == "topright" then
			frame.auraFrame:SetPoint("TOPLEFT", frame, "TOPRIGHT", 2, 0)
		elseif buffside == "bottomleft" then
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", w-size-2, 0)
		elseif buffside == "bottomright" then
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 2, 0)
		end
	else
		local side = ABF_db.profile[frame.group].auraSide
		if side == "bottom" then
			frame.auraFrame:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", -2, -2)
			frame.auraFrame:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT", 2, -2)
		elseif side == "top" then
			local split = PitBull:GetModule("Aura").db.profile[frame.group].split
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", -2, split and 18 or 2)
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 2, split and 18 or 2)
		elseif side == "left" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "LEFT", -2, 0)
		elseif side == "topright" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("TOPLEFT", frame, "TOPRIGHT", 2, 0)
		elseif side == "bottomright" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 2, 0)
		elseif side == "bottomleft" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", -2, 0)
		elseif side == "topleft" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("TOPRIGHT", frame, "TOPLEFT", -2, 0)
		elseif side == "right" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "RIGHT", 2, 0)
		end
	end
end
function ABF.auraFrame2(unit, frame)
	frame.auraFrame2:ClearAllPoints()
	local size = PitBull:GetModule("Aura").db.profile[frame.group].debuffsize
	local w = frame:GetWidth()
	frame.auraFrame2:SetHeight(size)
	frame.auraFrame2:SetWidth(w)
	local debuffside = ABF_db.profile[frame.group].debuffSide
	if debuffside == "bottom" then
		frame.auraFrame2:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", -2, -2)
	elseif debuffside == "top" then
		frame.auraFrame2:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", -2, 2)
	elseif debuffside == "left" then
		frame.auraFrame2:SetPoint("BOTTOMRIGHT", frame, "LEFT", w-size-2, -2)
	elseif debuffside == "right" then
		frame.auraFrame2:SetPoint("BOTTOMLEFT", frame, "RIGHT", 2, 0)
	elseif debuffside == "topleft" then
		frame.auraFrame2:SetPoint("TOPRIGHT", frame, "TOPLEFT", w-size-2, 0)
	elseif debuffside == "topright" then
		frame.auraFrame2:SetPoint("TOPLEFT", frame, "TOPRIGHT", 2, 0)
	elseif debuffside == "bottomleft" then
		frame.auraFrame2:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", w-size-2, 0)
	elseif debuffside == "bottomright" then
		frame.auraFrame2:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 2, 0)
	end
end
function ABF.portrait(unit, frame)
	local portraitSide = ABF_db.profile[frame.group].portraitSide
	local leftOrRight = portraitSide == "left" or portraitSide == "right"
	if portraitSide == "left" then
		frame.portrait:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0)
	elseif portraitSide == "right" then
		frame.portrait:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0)
	elseif portraitSide == "top" then
		frame.portrait:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0)
		frame.portrait:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0)
	else -- bottom then
		frame.portrait:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 0, 0)
		frame.portrait:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, 0)
	end
	
	if leftOrRight then
		local size
		if frame.powerBar then
			size = figureHeight('powerBar', frame) + figureHeight('healthBar', frame)
		else
			size = figureHeight('healthBar', frame)
		end
		frame.portrait:SetWidth(size)
		frame.portrait:SetHeight(size)
	else
		local size = figureHeight('portrait', frame)
		frame.portrait:SetHeight(size)
	end
end
function ABF.combatText(unit, frame)
	if frame.portrait then
		frame.combatText:SetPoint("CENTER", frame.portrait, "CENTER", 0, 0)
	else
		frame.combatText:SetPoint("CENTER", frame, "CENTER", 0, 0)
	end
end
function ABF.comboPoints(unit, frame)
	if frame.group ~= "target" then return end 
	local side = ABF_db.profile[frame.group].comboSide
	if side == "left" or side == "right" then
		if side == "left" then
			frame.comboPoints:SetPoint( "RIGHT", frame, "LEFT", 0, 0)
		else
			frame.comboPoints:SetPoint( "LEFT", frame, "RIGHT", 0, 0)
		end
		frame.comboPoints:SetWidth( 10 )
		frame.comboPoints:SetHeight( 50 )
		for i = 1, 5 do
			local combo = frame.comboPoints["combo"..i]
			combo:ClearAllPoints()
			if i == 1 then
				combo:SetPoint("BOTTOM", frame.comboPoints, "BOTTOM")
			else
				combo:SetPoint("BOTTOM", frame.comboPoints["combo"..(i-1)], "TOP", 0, 2)
			end			
		end
	elseif side == "top" or side == "bottom" then
		if side == "top" then
			frame.comboPoints:SetPoint( "TOPRIGHT", frame, "TOPRIGHT", 0, 0)
		else
			frame.comboPoints:SetPoint( "BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, 0)
		end
		frame.comboPoints:SetWidth( 50 )
		frame.comboPoints:SetHeight( 10 )
		for i = 1, 5 do
			local combo = frame.comboPoints["combo"..i]
			combo:ClearAllPoints()
			if i == 1 then
				combo:SetPoint("RIGHT", frame.comboPoints, "RIGHT")
			else
				combo:SetPoint("RIGHT", frame.comboPoints["combo"..(i-1)], "LEFT")
			end	
		end
	end
end
function ABF.combatIcon(unit, frame)
	frame.combatIcon:SetPoint("CENTER", frame, "BOTTOMLEFT", 0, 0)
	frame.combatIcon:SetWidth(24)
	frame.combatIcon:SetHeight(24)
end
function ABF.pvpIcon(unit, frame)
	frame.pvpIcon:SetPoint("CENTER", frame, "TOPRIGHT", 3, -5)
	frame.pvpIcon:SetWidth(27)
	frame.pvpIcon:SetHeight(27)
end
function ABF.restIcon(unit, frame)
	if frame.combatIcon then
		frame.restIcon:SetPoint("LEFT", frame.combatIcon, "RIGHT", 4, 0)
	else
		frame.restIcon:SetPoint("CENTER", frame, "BOTTOMLEFT", 0, 0)
	end
	frame.restIcon:SetWidth(24)
	frame.restIcon:SetHeight(24)
end
function ABF.leaderIcon(unit, frame)
	frame.leaderIcon:SetPoint("CENTER", frame, "TOPLEFT", 0, 0)
	frame.leaderIcon:SetWidth(18)
	frame.leaderIcon:SetHeight(18)
end
function ABF.masterIcon(unit, frame)
	if frame.leaderIcon then
		frame.masterIcon:SetPoint("LEFT", frame.leaderIcon, "RIGHT", 4, 0)
	else
		frame.masterIcon:SetPoint("CENTER", frame, "TOPLEFT", 0, 0)
	end
	frame.masterIcon:SetWidth(20)
	frame.masterIcon:SetHeight(20)
end

local function healthBarHeight_get(group)
	return ABF_db.profile[group].healthBarHeight
end
local function healthBarHeight_set(group, value)
	ABF_db.profile[group].healthBarHeight = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
function ABF.healthBar_options(group)
	return {
		name = L["Height"],
		desc = L["Relative height of the bar"],
		type = 'range',
		pass = true,
		passValue = group,
		get = healthBarHeight_get,
		set = healthBarHeight_set,
		min = 1,
		max = 12,
		step = 1,
	}
end

local function powerBarHeight_get(group)
	return ABF_db.profile[group].powerBarHeight
end
local function powerBarHeight_set(group, value)
	ABF_db.profile[group].powerBarHeight = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
function ABF.powerBar_options(group)
	return {
		name = L["Height"],
		desc = L["Relative height of the bar"],
		type = 'range',
		pass = true,
		passValue = group,
		get = powerBarHeight_get,
		set = powerBarHeight_set,
		min = 1,
		max = 12,
		step = 1,
	}
end

local function druidManaBarHeight_get(group)
	return ABF_db.profile[group].druidManaBarHeight
end
local function druidManaBarHeight_set(group, value)
	ABF_db.profile[group].druidManaBarHeight = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
function ABF.druidManaBar_options(group)
	if group == "player" and select(2, UnitClass("player")) == "DRUID" then
		return {
			name = L["Height Percentage"],
			desc = L["Percentage of height that the druid mana bar will take up of the normal power bar's height"],
			type = 'range',
			pass = true,
			passValue = group,
			get = druidManaBarHeight_get,
			set = druidManaBarHeight_set,
			min = 0.05,
			max = 0.95,
			step = 0.01,
			bigStep = 0.05,
			isPercent = true,
		}
	end
end

local function castBarHeight_get(group)
	return ABF_db.profile[group].castBarHeight
end
local function castBarHeight_set(group, value)
	ABF_db.profile[group].castBarHeight = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
function ABF.castBar_options(group)
	return {
		name = L["Height"],
		desc = L["Relative height of the bar"],
		type = 'range',
		pass = true,
		passValue = group,
		get = castBarHeight_get,
		set = castBarHeight_set,
		min = 1,
		max = 12,
		step = 1,
	}
end

local function expBarHeight_get(group)
	return ABF_db.profile[group].expBarHeight
end
local function expBarHeight_set(group, value)
	ABF_db.profile[group].expBarHeight = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
function ABF.expBar_options(group)
	if group == 'player' or group == 'pet' then
		return {
			name = L["Height"],
			desc = L["Relative height of the bar"],
			type = 'range',
			pass = true,
			passValue = group,
			get = expBarHeight_get,
			set = expBarHeight_set,
			min = 1,
			max = 12,
			step = 1,
		}
	end
end

local function repBarHeight_get(group)
	return ABF_db.profile[group].repBarHeight
end
local function repBarHeight_set(group, value)
	ABF_db.profile[group].repBarHeight = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
function ABF.repBar_options(group)
	if group == 'player' then
		return {
			name = L["Height"],
			desc = L["Relative height of the bar"],
			type = 'range',
			pass = true,
			passValue = group,
			get = repBarHeight_get,
			set = repBarHeight_set,
			min = 1,
			max = 12,
			step = 1,
		}
	end
end

local function portraitSide_get(group)
	return ABF_db.profile[group].portraitSide
end
local function portraitSide_set(group, value)
	ABF_db.profile[group].portraitSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
local function portraitSide_isLeftOrRight(group)
	local portraitSide = ABF_db.profile[group].portraitSide
	return portraitSide == "left" or portraitSide == "right"
end
local function portraitHeight_get(group)
	return ABF_db.profile[group].portraitHeight
end
local function portraitHeight_set(group, value)
	ABF_db.profile[group].portraitHeight = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
function ABF.portrait_options(group)
	return {
		name = L["Side"],
		desc = L["What side of the unit frame to position the portrait at."],
		type = "text",
		validate = {
			left = "Left",
			right = "Right",
			top = "Top",
		},
		get = portraitSide_get,
		set = portraitSide_set,
		passValue = group
	}, 	{
		name = L["Height"],
		desc = L["Relative height of the portrait"],
		type = 'range',
		pass = true,
		passValue = group,
		get = portraitHeight_get,
		set = portraitHeight_set,
		min = 1,
		max = 12,
		step = 1,
		disabled = portraitSide_isLeftOrRight,
	}
end

local function auraSide_get(group)
	return ABF_db.profile[group].auraSide
end
local function auraSide_set(group, value)
	ABF_db.profile[group].auraSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function buffSide_get(group)
	return ABF_db.profile[group].buffSide
end
local function buffSide_set(group, value)
	ABF_db.profile[group].buffSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function debuffSide_get(group)
	return ABF_db.profile[group].debuffSide
end
local function debuffSide_set(group, value)
	ABF_db.profile[group].debuffSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function splitOptions_isHidden(group)
	return not PitBull:GetModule("Aura").db.profile[group].split
end

local function nonsplitOptions_isHidden(group)
	return PitBull:GetModule("Aura").db.profile[group].split
end

function ABF.auraFrame_options(group)
	return {
		name = L["Side"],
		desc = L["Options to change the side which the auras are on."],
		type = "group",
		disabled = function()
			if not PitBull_Aura then PitBull_Aura = PitBull:GetModule("Aura") end
			if not PitBull_Aura or not PitBull_Aura.db or not PitBull_Aura.db.profile or not PitBull_Aura.db.profile[group] then return end
			return PitBull_Aura.db.profile[group].hidden
		end,
		args = {
			nonsplitside = {
				name = L["Side"],
				desc = L["Which side to position all aura icons at."],
				type = "text",
				validate = {
					left = "Left",
					topleft = "Top-left",
					bottomleft = "Bottom-left",
					right = "Right",
					topright = "Top-right",
					bottomright = "Bottom-right",
					top = "Top",
					bottom = "Bottom",
				},
				get = auraSide_get,
				set = auraSide_set,
				passValue = group,
				disabled = nonsplitOptions_isHidden,
			},
			buffside = {
				name = L["Buff side"],
				desc = L["Which side to position buff icons at."],
				type = "text",
				validate = {
					left = "Left",
					topleft = "Top-left",
					bottomleft = "Bottom-left",
					right = "Right",
					topright = "Top-right",
					bottomright = "Bottom-right",
					top = "Top",
					bottom = "Bottom",
				},
				get = buffSide_get,
				set = buffSide_set,
				passValue = group,
				disabled = splitOptions_isHidden,
			},
			debuffside = {
				name = L["Debuff side"],
				desc = L["Which side to position debuff icons at."],
				type = "text",
				validate = {
					left = "Left",
					topleft = "Top-left",
					bottomleft = "Bottom-left",
					right = "Right",
					topright = "Top-right",
					bottomright = "Bottom-right",
					top = "Top",
					bottom = "Bottom",
				},
				get = debuffSide_get,
				set = debuffSide_set,
				passValue = group,
				disabled = splitOptions_isHidden,
			},
		}
	}
end

local function comboSide_get(group)
	return ABF_db.profile[group].comboSide
end
local function comboSide_set(group, value)
	ABF_db.profile[group].comboSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

ABF.comboPoints_options_name = "Combo points"
ABF.comboPoints_options_desc = "Options for the combo point display."
function ABF.comboPoints_options(group)
	if group == "target" then
		return {
			name = L["Side"],
			desc = L["What side of the unit frame to position the combo point display at."],
			type = "text",
			validate = {
				left = "Left",
				right = "Right",
				top = "Top",
				bottom = "Bottom",
			},
			get = comboSide_get,
			set = comboSide_set,
			passValue = group,
		}
	end
end

ABF.extraFrames = {}

local classTextStyles = {
	["Standard"]            = "[Classification] [Level:DifficultyColor] [SmartClass:ClassColor] [DruidForm:Paren] [SmartRace]",
	["Hide"]                = "",
	["Custom"]              = "Custom",
	["Player Classes Only"] = "[Classification] [Level:DifficultyColor] [PlayerClass:ClassColor] [DruidForm:Paren] [SmartRace]",
	["Short"]               = "[DifficultyColor][Level][Plus][White] [SmartRace]",
}
function ABF.extraFrames.classText(unit, frame)
	local classText = newFrame("FontString", frame.overlay, "OVERLAY")
	frame.classText = classText
	local font, fontsize = PitBull:GetFont()
	classText:SetFont( font, fontsize )
	classText:SetShadowColor(0, 0, 0, 1)
	classText:SetShadowOffset(0.8, -0.8)
	classText:SetNonSpaceWrap(false)
	classText:SetJustifyH("LEFT")
	local code
	if ABF_db.profile[frame.group].classTextStyle == "Custom" then
		code = ABF_db.profile[frame.group].classTextCustom
	else
		code = classTextStyles[ABF_db.profile[frame.group].classTextStyle]
	end
	DogTag:RegisterFontString(classText, frame, unit, code)
end

function ABF.classText(unit, frame)
	if frame.powerBar then
		frame.classText:SetPoint("LEFT", frame.powerBar, "LEFT", 3, 0)
		if frame.powerText then
			frame.classText:SetPoint("RIGHT", frame.powerText, "LEFT", -5, 0)
		end
	end
end

local function getTextStyle(group)
	return ABF_db.profile[group].classTextStyle
end
local function setTextStyle(group, value)
	ABF_db.profile[group].classTextStyle = value
	
	local code
	if value == "Custom" then
		code = ABF_db.profile[group].classTextCustom
	else
		code = classTextStyles[value]
		ABF_db.profile[group].classTextCustom = code
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.classText, frame, unit, code)
	end
end
local function getTextCustom(group)
	return ABF_db.profile[group].classTextCustom
end
local function setTextCustom(group, value)
	value = DogTag:FixCasing(value)
	ABF_db.profile[group].classTextCustom = value
	
	if ABF_db.profile[group].classTextStyle ~= "Custom" then
		return
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.classText, frame, unit, value)
	end
end
ABF.classText_options_name = "Class text"
ABF.classText_options_desc = "Customize the class text."
function ABF.classText_options(group)
	local textValidate = {}
	for k in pairs(classTextStyles) do
		textValidate[k] = k
	end
	return {
		type = "text",
		name = L["Style"],
		desc = L["Use a preconfigured style."],
		validate = textValidate,
		validateDesc = classTextStyles,
		get = getTextStyle,
		set = setTextStyle,
		passValue = group,
	}, {
		type = "text",
		name = L["Custom"],
		desc = L["Create a custom style for this unit type."],
		get = getTextCustom,
		set = setTextCustom,
		usage = "DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0",
		passValue = group,
	}
end

local healthTextStyles = {
	["Absolute"]       = "[Status:SureHP:PercentHP:Percent]",
	["Absolute Short"] = "[Status:SureHP:Short:PercentHP:Percent]",
	["Difference"]     = "[Status:SmartMissingHP]",
	["Percent"]        = "[Status:PercentHP:Percent]",
	["Hide"]           = "",
	["Mini"]           = "[CurHP:VeryShort]",
	["Custom"]         = "Custom",
	["Smart"]          = "[IsFriend?Status:MissingHP:HideZero:Short:Color(ff7f7f)!Status:SmartHP:Short]",
	["Absolute and Percent"]  = "[~Status?FractionalHP:Short] || [Status:PercentHP:Percent]",
	["Informational"]  = "[IsFriend?~Status?MissingHP:HideZero:Short:Color(ff7f7f)] || [~Status?FractionalHP:Short] || [Status:PercentHP:Percent]"
}

function ABF.extraFrames.healthText(unit, frame)
	local healthText = newFrame("FontString", frame.overlay, "OVERLAY")
	frame.healthText = healthText
	local font, fontsize = PitBull:GetFont()
	healthText:SetFont(font, fontsize)
	healthText:SetShadowColor(0, 0, 0, 1)
	healthText:SetShadowOffset(0.8, -0.8)
	healthText:SetNonSpaceWrap(false)
	healthText:SetJustifyH("LEFT")
	local code
	if ABF_db.profile[frame.group].healthTextStyle == "Custom" then
		code = ABF_db.profile[frame.group].healthTextCustom
	else
		code = healthTextStyles[ABF_db.profile[frame.group].healthTextStyle]
	end
	DogTag:RegisterFontString(healthText, frame, unit, code)
end

function ABF.healthText(unit, frame)
	if frame.healthBar then
		frame.healthText:SetPoint("RIGHT", frame.healthBar, "RIGHT", -3, 0)
	end
end

local function getTextStyle(group)
	return ABF_db.profile[group].healthTextStyle
end
local function setTextStyle(group, value)
	ABF_db.profile[group].healthTextStyle = value
	
	local code
	if value == "Custom" then
		code = ABF_db.profile[group].healthTextCustom
	else
		code = healthTextStyles[value]
		ABF_db.profile[group].healthTextCustom = code
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.healthText, frame, unit, code)
	end
end
local function getTextCustom(group)
	return ABF_db.profile[group].healthTextCustom
end
local function setTextCustom(group, value)
	value = DogTag:FixCasing(value)
	ABF_db.profile[group].healthTextCustom = value
	
	if ABF_db.profile[group].healthTextStyle ~= "Custom" then
		return
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.healthText, frame, unit, value)
	end
end

ABF.healthText_options_name = "Health text"
ABF.healthText_options_desc = "Customize the health text."
function ABF.healthText_options(group)
	local textValidate = {}
	for k in pairs(healthTextStyles) do
		textValidate[k] = k
	end
	return {
		type = "text",
		name = L["Style"],
		desc = L["Use a preconfigured style."],
		validate = textValidate,
		validateDesc = healthTextStyles,
		get = getTextStyle,
		set = setTextStyle,
		passValue = group,
	}, {
		type = "text",
		name = L["Custom"],
		desc = L["Create a custom style for this unit type."],
		get = getTextCustom,
		set = setTextCustom,
		usage = "DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0",
		passValue = group,
	}
end

local nameTextStyles = {
	["Standard"]             = "[Name] [AFKDND:Angle]",
	["Hostility Colored"]    = "[Name:HostileColor] [AFKDND:Angle]",
	["Hide"]                 = "",
	["Custom"]               = "Custom",
	["Class Colored"]        = "[Name:ClassColor] [AFKDND:Angle]",
	["Long"]                 = "[Level] [Name:ClassColor] [AFKDND:Angle]",
	["Long w/ Druid form"]   = "[Level] [Name:ClassColor] [DruidForm:Paren] [AFKDND:Angle]",
}
function ABF.extraFrames.nameText(unit, frame)
	local nameText = newFrame("FontString", frame.overlay, "OVERLAY")
	frame.nameText = nameText
	local font, fontsize = PitBull:GetFont()
	nameText:SetFont(font, fontsize)
	nameText:SetShadowColor(0, 0, 0, 1)
	nameText:SetShadowOffset(0.8, -0.8)
	nameText:SetNonSpaceWrap(false)
	nameText:SetJustifyH("LEFT")
	
	local code
	if ABF_db.profile[frame.group].nameTextStyle == "Custom" then
		code = ABF_db.profile[frame.group].nameTextCustom
	else
		code = nameTextStyles[ABF_db.profile[frame.group].nameTextStyle]
	end
	DogTag:RegisterFontString(nameText, frame, unit, code)
end

function ABF.nameText(unit, frame)
	if frame.healthBar then
		frame.nameText:SetPoint("LEFT", frame.healthBar, "LEFT", 3, 0)
		if frame.healthText then
			frame.nameText:SetPoint("RIGHT", frame.healthText, "LEFT", -5, 0)
		end
	end
end

local function getTextStyle(group)
	return ABF_db.profile[group].nameTextStyle
end
local function setTextStyle(group, value)
	ABF_db.profile[group].nameTextStyle = value
	
	local code
	if value == "Custom" then
		code = ABF_db.profile[group].nameTextCustom
	else
		code = nameTextStyles[value]
		ABF_db.profile[group].nameTextCustom = code
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.nameText, frame, unit, code)
	end
end
local function getTextCustom(group)
	return ABF_db.profile[group].nameTextCustom
end
local function setTextCustom(group, value)
	value = DogTag:FixCasing(value)
	ABF_db.profile[group].nameTextCustom = value
	
	if ABF_db.profile[group].nameTextStyle ~= "Custom" then
		return
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.nameText, frame, unit, value)
	end
end

ABF.nameText_options_name = "Name text"
ABF.nameText_options_desc = "Customize the name text."
function ABF.nameText_options(group)
	local textValidate = {}
	for k in pairs(nameTextStyles) do
		textValidate[k] = k
	end
	return {
		type = "text",
		name = L["Style"],
		desc = L["Use a preconfigured style."],
		validate = textValidate,
		validateDesc = nameTextStyles,
		get = getTextStyle,
		set = setTextStyle,
		passValue = group,
	}, {
		type = "text",
		name = L["Custom"],
		desc = L["Create a custom style for this unit type."],
		get = getTextCustom,
		set = setTextCustom,
		usage = "DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0",
		passValue = group,
	}
end

local powerTextStyles = {
	["Absolute"]       = "[~HasNoMP?FractionalMP]",
	["Absolute Short"] = "[~HasNoMP?FractionalMP:Short]",
	["Difference"]     = "[MissingMP:Negate]",
	["Percent"]        = "[PercentMP:Percent]",
	["Mini"]           = "[~HasNoMP?CurMP:VeryShort]",
	["Hide"]           = "",
	["Custom"]         = "Custom",
	["Smart"]          = "[MissingMP:HideZero:Short:Color(7f7fff)]",
}
function ABF.extraFrames.powerText(unit, frame)
	local powerText = newFrame("FontString", frame.overlay, "OVERLAY")
	frame.powerText = powerText
	local font, fontsize = PitBull:GetFont()
	powerText:SetFont(font, fontsize)
	powerText:SetShadowColor(0, 0, 0, 1)
	powerText:SetShadowOffset(0.8, -0.8)
	powerText:SetNonSpaceWrap(false)
	powerText:SetJustifyH("LEFT")
	local code
	if ABF_db.profile[frame.group].powerTextStyle == "Custom" then
		code = ABF_db.profile[frame.group].powerTextCustom
	else
		code = powerTextStyles[ABF_db.profile[frame.group].powerTextStyle]
	end
	DogTag:RegisterFontString(powerText, frame, unit, code)
end

function ABF.powerText(unit, frame)
	if frame.powerBar then
		frame.powerText:SetPoint("RIGHT", frame.powerBar, "RIGHT", -3, 0)
	end
end

local function getTextStyle(group)
	return ABF_db.profile[group].powerTextStyle
end
local function setTextStyle(group, value)
	ABF_db.profile[group].powerTextStyle = value
	
	local code
	if value == "Custom" then
		code = ABF_db.profile[group].powerTextCustom
	else
		code = powerTextStyles[value]
		ABF_db.profile[group].powerTextCustom = code
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.powerText, frame, unit, code)
	end
end
local function getTextCustom(group)
	return ABF_db.profile[group].powerTextCustom
end
local function setTextCustom(group, value)
	value = DogTag:FixCasing(value)
	ABF_db.profile[group].powerTextCustom = value
	
	if ABF_db.profile[group].powerTextStyle ~= "Custom" then
		return
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.powerText, frame, unit, value)
	end
end

ABF.powerText_options_name = "Power text"
ABF.powerText_options_desc = "Customize the power text."
function ABF.powerText_options(group)
	local textValidate = {}
	for k in pairs(powerTextStyles) do
		textValidate[k] = k
	end
	return {
		type = "text",
		name = L["Style"],
		desc = L["Use a preconfigured style."],
		validate = textValidate,
		validateDesc = powerTextStyles,
		get = getTextStyle,
		set = setTextStyle,
		passValue = group,
	}, {
		type = "text",
		name = L["Custom"],
		desc = L["Create a custom style for this unit type."],
		get = getTextCustom,
		set = setTextCustom,
		usage = "DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0",
		passValue = group,
	}
end

local comboTextStyles = {
	["Standard"]       = "[IsEnergy#player?Combos:HideZero]",
	["Hide"]           = "",
	["Custom"]         = "Custom",
}
function ABF.extraFrames.comboText(unit, frame)
	if frame.group ~= "target" then return end 
	local comboText = newFrame("FontString", frame.overlay, "OVERLAY")
	frame.comboText = comboText
	local font, fontsize = PitBull:GetFont()
	comboText:SetFont(font, fontsize)
	comboText:SetShadowColor(0, 0, 0, 1)
	comboText:SetShadowOffset(0.8, -0.8)
	comboText:SetNonSpaceWrap(false)
	comboText:SetJustifyH("LEFT")
	local code
	if ABF_db.profile[frame.group].comboTextStyle == "Custom" then
		code = ABF_db.profile[frame.group].comboTextCustom
	else
		code = comboTextStyles[ABF_db.profile[frame.group].comboTextStyle]
	end
	DogTag:RegisterFontString(comboText, frame, unit, code)
end

function ABF.comboText(unit, frame)
	if frame.group ~= "target" then return end
	if not frame.powerBar then
		frame.comboText:Hide()
	else
		frame.comboText:SetPoint("CENTER", frame.powerBar, "CENTER")
	end
end

local function getTextStyle(group)
	return ABF_db.profile[group].comboTextStyle
end
local function setTextStyle(group, value)
	ABF_db.profile[group].comboTextStyle = value
	
	local code
	if value == "Custom" then
		code = ABF_db.profile[group].comboTextCustom
	else
		code = comboTextStyles[value]
		ABF_db.profile[group].comboTextCustom = code
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup("target") do
		DogTag:RegisterFontString(frame.comboText, frame, unit, code)
	end
end
local function getTextCustom(group)
	return ABF_db.profile[group].comboTextCustom
end
local function setTextCustom(group, value)
	ABF_db.profile[group].comboTextCustom = value
	
	if ABF_db.profile[group].comboTextStyle ~= "Custom" then
		return
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup("target") do
		DogTag:RegisterFontString(frame.comboText, frame, unit, value)
	end
end

ABF.comboText_options_name = "Combo text"
ABF.comboText_options_desc = "Customize the combo points text."
function ABF.comboText_options(group)
	local textValidate = {}
	for k in pairs(comboTextStyles) do
		textValidate[k] = k
	end
	return {
		type = "text",
		name = L["Style"],
		desc = L["Use a preconfigured style."],
		validate = textValidate,
		validateDesc = comboTextStyles,
		get = getTextStyle,
		set = setTextStyle,
		passValue = group,
	}, {
		type = "text",
		name = L["Custom"],
		desc = L["Create a custom style."],
		get = getTextCustom,
		set = setTextCustom,
		usage = "DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0",
		passValue = group,
	}
end

local hasPetExp = select(2, UnitClass("player")) == "HUNTER"
local experienceTextStyles = {
	["Standard"]       = "[FractionalXP] [PercentXP:Percent:Paren] [PercentRestXP:HideZero:Percent:Prepend(R: )]",
	["On Mouse-over"]       = "[IsMouseover?FractionalXP] [IsMouseover?PercentXP:Percent:Paren] [IsMouseover?PercentRestXP:HideZero:Percent:Prepend(R: )]",
	["Hide"]           = "",
	["Custom"]         = "Custom",
}
function ABF.extraFrames.experienceText(unit, frame)
	if frame.group ~= "player" and (not hasPetExp or frame.group ~= "pet") then
		return
	end
	if not frame.expBar then
		return
	end
	local experienceText = newFrame("FontString", frame.overlay, "OVERLAY")
	frame.experienceText = experienceText
	local font, fontsize = PitBull:GetFont()
	experienceText:SetFont(font, fontsize)
	experienceText:SetShadowColor(0, 0, 0, 1)
	experienceText:SetShadowOffset(0.8, -0.8)
	experienceText:SetNonSpaceWrap(false)
	experienceText:SetJustifyH("CENTER")
	local code
	if ABF_db.profile[frame.group].experienceTextStyle == "Custom" then
		code = ABF_db.profile[frame.group].experienceTextCustom
	else
		code = experienceTextStyles[ABF_db.profile[frame.group].experienceTextStyle]
	end
	DogTag:RegisterFontString(experienceText, frame, unit, code)
end

function ABF.experienceText(unit, frame)
	frame.experienceText:SetAllPoints(frame.expBar)
end

local function getTextStyle(group)
	return ABF_db.profile[group].experienceTextStyle
end
local function setTextStyle(group, value)
	ABF_db.profile[group].experienceTextStyle = value
	
	local code
	if value == "Custom" then
		code = ABF_db.profile[group].experienceTextCustom
	else
		code = experienceTextStyles[value]
		ABF_db.profile[group].experienceTextCustom = code
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.experienceText, frame, unit, code)
	end
end
local function getTextCustom(group)
	return ABF_db.profile[group].experienceTextCustom
end
local function setTextCustom(group, value)
	ABF_db.profile[group].experienceTextCustom = value
	
	if ABF_db.profile[group].experienceTextStyle ~= "Custom" then
		return
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.experienceText, frame, unit, value)
	end
end

ABF.experienceText_options_name = "Experience text"
ABF.experienceText_options_desc = "Customize the experience text."
function ABF.experienceText_options(group)
	local textValidate = {}
	for k in pairs(experienceTextStyles) do
		textValidate[k] = k
	end
	return {
		type = "text",
		name = L["Style"],
		desc = L["Use a preconfigured style."],
		validate = textValidate,
		validateDesc = experienceTextStyles,
		get = getTextStyle,
		set = setTextStyle,
		passValue = group,
	}, {
		type = "text",
		name = L["Custom"],
		desc = L["Create a custom style."],
		get = getTextCustom,
		set = setTextCustom,
		usage = "DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0",
		passValue = group,
	}
end

local repTextStyles = {
	["Standard"]       = "[~IsMouseOver?FractionalRep] [~IsMouseOver?PercentRep:Percent:Paren!RepName]",
	["Hide"]           = "",
	["Custom"]         = "Custom",
}
function ABF.extraFrames.repText(unit, frame)
	if frame.group ~= "player" then
		return
	end
	if not frame.repBar then
		return
	end	
	local repText = newFrame("FontString", frame.overlay, "OVERLAY")
	frame.repText = repText
	repText:SetFont(PitBull:GetFont())
	repText:SetShadowColor(0, 0, 0, 1)
	repText:SetShadowOffset(0.8, -0.8)
	repText:SetNonSpaceWrap(false)
	repText:SetJustifyH("CENTER")
	local code
	if ABF_db.profile[frame.group].repTextStyle == "Custom" then
		code = ABF_db.profile[frame.group].repTextCustom
	else
		code = repTextStyles[ABF_db.profile[frame.group].repTextStyle]
	end
	DogTag:RegisterFontString(repText, frame, unit, code)
end

function ABF.repText(unit, frame)
	frame.repText:SetAllPoints(frame.repBar)
end

local function getTextStyle(group)
	return ABF_db.profile[group].repTextStyle
end
local function setTextStyle(group, value)
	ABF_db.profile[group].repTextStyle = value
	
	local code
	if value == "Custom" then
		code = ABF_db.profile[group].repTextCustom
	else
		code = repTextStyles[value]
		ABF_db.profile[group].repTextCustom = code
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.repText, frame, unit, code)
	end
end
local function getTextCustom(group)
	return ABF_db.profile[group].repTextCustom
end
local function setTextCustom(group, value)
	ABF_db.profile[group].repTextCustom = value
	
	if ABF_db.profile[group].repTextStyle ~= "Custom" then
		return
	end
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		DogTag:RegisterFontString(frame.repText, frame, unit, value)
	end
end

ABF.repText_options_name = "Reputation text"
ABF.repText_options_desc = "Customize the reputation text."
function ABF.repText_options(group)
	local textValidate = {}
	for k in pairs(repTextStyles) do
		textValidate[k] = k
	end
	return {
		type = "text",
		name = L["Style"],
		desc = L["Use a preconfigured style."],
		validate = textValidate,
		validateDesc = repTextStyles,
		get = getTextStyle,
		set = setTextStyle,
		passValue = group,
	}, {
		type = "text",
		name = L["Custom"],
		desc = L["Create a custom style."],
		get = getTextCustom,
		set = setTextCustom,
		usage = "DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0",
		passValue = group,
	}
end

PitBull:RegisterLayout("ABF", ABF)
