--[[
Name: DogTag-1.0
Revision: $Rev: 35496 $
Author: Cameron Kenneth Knight (ckknight@gmail.com)
Inspired By: WatchDog by Vika
Website: http://www.wowace.com/
Documentation: http://www.wowace.com/wiki/DogTag-1.0
SVN: svn://svn.wowace.com/wowace/trunk/DogTag-1.0/DogTag-1.0
Description: A library to provide a markup syntax for unit frame FontStrings
Dependencies: AceLibrary
License: LGPL v2.1
]]

-------------------------------------------------------------------
-- FOR TAG INFORMATION SEE http://www.wowace.com/wiki/DogTag-1.0 --
-------------------------------------------------------------------

local MAJOR_VERSION = "DogTag-1.0"
local MINOR_VERSION = tonumber(("$Revision: 35496 $"):match("%d+")) or 0

-- This ensures the code is only executed if the libary doesn't already exist, or is a newer version
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local DogTag = {}

local L = {
	-- races
	["Blood Elf"] = "Blood Elf",
	["Draenei"] = "Draenei",
	["Dwarf"] = "Dwarf",
	["Gnome"] = "Gnome",
	["Human"] = "Human",
	["Night Elf"] = "Night Elf",
	["Orc"] = "Orc",
	["Tauren"] = "Tauren",
	["Troll"] = "Troll",
	["Undead"] = "Undead",
	
	-- short races
	["BE"] = "BE",
	["Dr"] = "Dr",
	["Dw"] = "Dw",
	["Gn"] = "Gn",
	["Hu"] = "Hu",
	["NE"] = "NE",
	["Or"] = "Or",
	["Ta"] = "Ta",
	["Tr"] = "Tr",
	["Ud"] = "Ud",
	
	-- classes
	["Warrior"] = "Warrior",
	["Priest"] = "Priest",
	["Mage"] = "Mage",
	["Shaman"] = "Shaman",
	["Paladin"] = "Paladin",
	["Warlock"] = "Warlock",
	["Druid"] = "Druid",
	["Rogue"] = "Rogue",
	["Hunter"] = "Hunter",
	
	-- short classes
	["Wr"] = "Wr",
	["Pr"] = "Pr",
	["Ma"] = "Ma",
	["Sh"] = "Sh",
	["Pa"] = "Pa",
	["Wl"] = "Wl",
	["Dr"] = "Dr",
	["Ro"] = "Ro",
	["Hu"] = "Hu",
	
	["Player"] = "Player",
	["Target"] = "Target",
	["Focus"] = "Focus",
	["Mouse-over"] = "Mouse-over",
	["%s's pet"] = "%s's pet",
	["%s's target"] = "%s's target",
	["Party member #%d"] = "Party member #%d",
	["Raid member #%d"] = "Raid member #%d",
	
	-- classifications
	["Rare"] = ITEM_QUALITY3_DESC,
	["Rare-Elite"] = ITEM_QUALITY3_DESC .. "-" .. ELITE,
	["Elite"] = ELITE,
	["Boss"] = BOSS,
	-- short classifications
	["r"] = "r",
	["r+"] = "r+",
	["+"] = "+",
	["b"] = "b",
	
	["Soulstone"] = "Soulstone",
	["Dead"] = DEAD,
	["Feigned Death"] = "Feigned Death",
	["Ghost"] = "Ghost",
	["Offline"] = PLAYER_OFFLINE,
	["Combat"] = "Combat",
	["Resting"] = "Resting",
	["Tapped"] = "Tapped",
	["AFK"] = "AFK",
	["DND"] = "DND",
	
	["True"] = "True",
	
	["Rage"] = RAGE,
	["Focus"] = FOCUS,
	["Energy"] = ENERGY,
	["Mana"] = MANA,
	
	["PvP"] = PVP,
	
	-- genders
	["Male"] = MALE,
	["Female"] = FEMALE,
	
	-- forms
	["Bear"] = "Bear",
	["Cat"] = "Cat",
	["Moonkin"] = "Moonkin",
	["Aquatic"] = "Aquatic",
	["Flight"] = "Flight",
	["Travel"] = "Travel",
	["Tree"] = "Tree",
	
	-- shortgenders
	["m"] = "m",
	["f"] = "f",
	
	["Leader"] = "Leader",
}
if GetLocale() == "deDE" then
	-- alter L here
end

setmetatable(L, {__index = function(self, key) error(("Error indexing L[%q]"):format(tostring(key)), 2) end})

local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local UNKNOWN = UNKNOWN or "Unknown"
local unpack = unpack
local pcall = pcall
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local colors = {}
for class, data in pairs(RAID_CLASS_COLORS) do
	colors[class] = { data.r, data.g, data.b, }
end
colors.unknown = { 0.8, 0.8, 0.8 }
colors.hostile = { 226/255, 45/255, 75/255 }
colors.neutral = { 1, 1, 34/255 }
colors.friendly = { 0.2, 0.8, 0.15 }
colors.civilian = { 48/255, 113/255, 191/255 }
colors.tapped = { 0.5, 0.5, 0.5 }
colors.dead = { 0.6, 0.6, 0.6 }
colors.disconnected = { 0.7, 0.7, 0.7 }
colors.inCombat = { 1, 0, 0 }
colors.petHappy = { 0, 1, 0 }
colors.petNeutral = { 1, 1, 0 }
colors.petAngry = { 1, 0, 0 }
colors.rage = { 226/255, 45/255, 75/255 }
colors.energy = { 1, 220/255, 25/255 }
colors.focus = { 1, 210/255, 0 }
colors.mana = { 48/255, 113/255, 191/255 }
colors.minHP = { 1, 0, 0 }
colors.midHP = { 1, 1, 0 }
colors.maxHP = { 0, 1, 0 }

local portrait = CreateFrame("PlayerModel")
local lastMouseFocus

local IsLegitimateUnit = {player=true,pet=true,target=true,focus=true,mouseover=true}
for i = 1, 4 do
	IsLegitimateUnit['party' .. i] = true
	IsLegitimateUnit['partypet' .. i] = true
end
for i = 1, 40 do
	IsLegitimateUnit['raid' .. i] = true
	IsLegitimateUnit['raidpet' .. i] = true
end
setmetatable(IsLegitimateUnit, {__index=function(self, unit)
	if not unit:find("target$") then
		self[unit] = false
		return false
	end
	local nonTarget = unit:sub(1, -7)
	local good = self[nonTarget]
	self[unit] = good
	return good
end})

local currentAuras = setmetatable({},{__index=function(self, unit)
	local t = new()
	for i = 1, 40 do
		local _,_,texture = UnitBuff(unit, i)
		if not texture then
			break
		end
		t[texture] = true
	end
	for i = 1, 40 do
		local _,_,texture = UnitDebuff(unit, i)
		if not texture then
			break
		end
		t[texture] = true
	end
	self[unit] = t
	return t
end})

local Modifiers; Modifiers = {}
-- when writing modifiers, inputs can be a number or a string. Also, any of these can be returned, as well as nil, which means a blank string.
-- the modifier itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
-- note: can take an argument, either nil, a number, a string.
Modifiers.IsEqual = {"if value ~= ${arg} then value = nil end", value = "number;string", arg = "number;string", ret = "same;nil"}

Modifiers["~IsEqual"] = {"if value == ${arg} then value = nil end", value = "number;string", arg = "number;string", ret = "same;nil"}

Modifiers.IsLess = {"if type(value) ~= ${argtype} then if value >= ${arg} then value = nil end else if tostring(value) >= ${arg:string} then value = nil end end", value = "number;string", arg = "number;string", ret = "same;nil", globals = "tostring"}

Modifiers["~IsLess"] = {"if type(value) ~= ${argtype} then if value < ${arg} then value = nil end else if tostring(value) < ${arg:string} then value = nil end end", value = "number;string", arg = "number;string", ret = "same;nil"}

Modifiers.IsLessEqual = {"if type(value) ~= ${argtype} then if value > ${arg} then value = nil end else if tostring(value) > ${arg:string} then value = nil end end", value = "number;string", arg = "number;string", ret = "same;nil", globals = "tostring"}

Modifiers["~IsLessEqual"] = {"if type(value) ~= ${argtype} then if value <= ${arg} then value = nil end else if tostring(value) <= ${arg:string} then value = nil end end", value = "number;string", arg = "number;string", ret = "same;nil", globals = "tostring"}

Modifiers.IsGreater = "~IsLessEqual"
Modifiers.IsGreaterEqual = "~IsLess"
Modifiers.IsNotEqual = "~IsEqual"

Modifiers.Hide = "~IsEqual"
Modifiers.HideLess = "~IsLess"
Modifiers.HideGreater = "~IsGreater"
Modifiers.HideLessEqual = "~IsLessEqual"
Modifiers.HideGreaterEqual = "~IsGreaterEqual"
Modifiers.HideUnless = "~IsNotEqual"
Modifiers.HideZero = "Hide(0)"

Modifiers.Prepend = {"value = ${arg} .. value", value = "number;string", arg = "string", ret = "string"}

Modifiers.Append = {"value = value .. ${arg}", value = "number;string", arg = "string", ret = "string"}

Modifiers.Negate = {"value = -value", value = "number", ret = "number"}

Modifiers.Percent = {"value = value .. '%'", value = "number", ret = "string"}

Modifiers.Short = {[[
	if type(value) ~= "number" then
		local a,b = value:match("^(%d+)/(%d+)")
		if a then
			a, b = tonumber(a), tonumber(b)
			if b >= 10000000 then
				b = ("%.1fm"):format(b / 1000000)
			elseif b >= 1000000 then
				b = ("%.2fm"):format(b / 1000000)
			elseif b >= 100000 then
				b = ("%.0fk"):format(b / 1000)
			elseif b >= 10000 then
				b = ("%.1fk"):format(b / 1000)
			end
			if a >= 10000000 then
				a = ("%.1fm"):format(a / 1000000)
			elseif a >= 1000000 then
				a = ("%.2fm"):format(a / 1000000)
			elseif a >= 100000 then
				a = ("%.0fk"):format(a / 1000)
			elseif a >= 10000 then
				a = ("%.1fk"):format(a / 1000)
			end
			value = a.."/"..b
		end
	else	
		if value >= 10000000 then
			value = ("%.1fm"):format(value / 1000000)
		elseif value >= 1000000 then
			value = ("%.2fm"):format(value / 1000000)
		elseif value >= 100000 then
			value = ("%.0fk"):format(value / 1000)
		elseif value >= 10000 then
			value = ("%.1fk"):format(value / 1000)
		end
	end
]],	value = "number;string", ret = "same;nil"}

Modifiers.VeryShort = {[[
	if type(value) ~= "number" then
		local a,b = value:match("^(%d+)/(%d+)")
		if a then
			a, b = tonumber(a), tonumber(b)
			if b >= 1000000 then
				b = ("%.0fm"):format(b / 1000000)
			elseif b >= 1000 then
				b = ("%.0fk"):format(b / 1000)
			end
			if a >= 1000000 then
				a = ("%.0fm"):format(a / 1000000)
			elseif a >= 1000 then
				a = ("%.0fk"):format(a / 1000)
			end
			value = a.."/"..b
		end
	else	
		if value >= 1000000 then
			value = ("%.0fm"):format(value / 1000000)
		elseif value >= 1000 then
			value = ("%.0fk"):format(value / 1000)
		end
	end
]], value = "number;string", ret = "same;nil"}

Modifiers.Upper = {"value = value:upper()", value = "string", ret = "string"}

Modifiers.Lower = {"value = value:lower()", value = "string", ret = "string"}

Modifiers.Bracket = {"value = '[' .. value .. ']'", value = "number;string", ret = "string"}

Modifiers.Paren = {"value = '(' .. value .. ')'", value = "number;string", ret = "string"}

Modifiers.Angle = {"value = '<' .. value .. '>'", value = "number;string", ret = "string"}

Modifiers.Brace = {"value = '{' .. value .. '}'", value = "number;string", ret = "string"}

Modifiers.Trunc = {[[do
	local len = 0
	for i = 1, ${arg} do
		local b = value:byte(len+1)
		if not b then
			break
		elseif b <= 127 then
			len = len + 1
		elseif b <= 223 then
			len = len + 2
		elseif b <= 239 then
			len = len + 3
		else
			len = len + 4
		end
	end
	value = value:sub(1, len)
end]], arg = "number", value = "string", ret = "string"}

Modifiers.Contains = {"if not value:find(${arg}) then value = nil end", value = "string", arg = "string", ret = "string;nil"}

Modifiers["~Contains"] = {"if value:find(${arg}) then value = nil end", value = "string", arg = "string", ret = "string;nil"}

Modifiers.HideContains = "~Contains"

Modifiers.ShortClass = {([[
	if value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	else
		value = nil
	end]]):format(L["Priest"], L["Pr"], L["Mage"], L["Ma"], L["Shaman"], L["Sh"], L["Paladin"], L["Pa"], L["Warlock"], L["Wl"], L["Druid"], L["Dr"], L["Rogue"], L["Ro"], L["Hunter"], L["Hu"], L["Warrior"], L["Wr"]), ret = "string;nil"}

Modifiers.ShortRace = {([[
	if value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	elseif value == %q then
		value = %q
	else
		value = nil
	end]]):format(L["Blood Elf"], L["BE"], L["Draenei"], L["Dr"], L["Dwarf"], L["Dw"], L["Gnome"], L["Gn"], L["Human"], L["Hu"], L["Night Elf"], L["NE"], L["Orc"], L["Or"], L["Tauren"], L["Ta"], L["Troll"], L["Tr"], L["Undead"], L["Ud"]), ret = "string;nil" }

Modifiers.RepColor = {function(data)
	return [[do
		local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
		if not repname then
		]] .. (data.isMod and [[
			value = "|r" .. value
		]] or [[
			value = "|r"
		]]) .. [[
		else
			local info = FACTION_BAR_COLORS[repreaction]
		]] .. (data.isMod and [[
			value = ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
		]] or [[
			value = ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
		]]) .. [[
		end
	end]]
end, events = "Reputation", globals = "GetWatchedFactionInfo;FACTION_BAR_COLORS"}

Modifiers.StatusColor = {function(data)
	return [[do
		local r,g,b
		if not UnitIsConnected(${unit}) then
			r,g,b = unpack(colors.disconnected)
		elseif UnitIsDeadOrGhost(${unit}) then
			r,g,b = unpack(colors.dead)
		else
		]] .. (data.isMod and [[
			value = "|r" .. value
		]] or [[
			value = "|r"
		]]) .. [[
		end
		if r then
		]] .. (data.isMod and [[
			value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
		]] or [[
			value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
		]]) .. [[
		end
	end]]
end, value = "number;string", ret = "string", events = "PLAYER_ALIVE;PLAYER_DEAD;PLAYER_UNGHOST", globals = "unpack;UnitIsConnected;UnitIsDeadOrGhost"}

Modifiers.HappyColor = {function(data)
	return [[do
		local x = GetPetHappiness()
		local r,g,b
		if x == 3 then
			r,g,b = unpack(colors.petHappy)
		elseif x == 2 then
			r,g,b = unpack(colors.petNeutral)
		elseif x == 1 then
			r,g,b = unpack(colors.petAngry)
		else
		]] .. (data.isMod and [[
			value = "|r" .. value
		]] or [[
			value = "|r"
		]]) .. [[
		end
		if r then
		]] .. (data.isMod and [[
			value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
		]] or [[
			value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
		]]) .. [[
		end
	end]]
end, value = "number;string", ret = "string", events = "UNIT_HAPPINESS", globals = "unpack;GetPetHappiness" }

Modifiers.HostileColor = {function(data)
	return [[do
		local r, g, b
		
		if UnitIsPlayer(${unit}) or UnitPlayerControlled(${unit}) then
			if UnitCanAttack(${unit}, "player") then
				-- they can attack me
				if UnitCanAttack("player", ${unit}) then
					-- and I can attack them
					r, g, b = unpack(colors.hostile)
				else
					-- but I can't attack them
					r, g, b = unpack(colors.civilian)
				end
			elseif UnitCanAttack("player", ${unit}) then
				-- they can't attack me, but I can attack them
				r, g, b = unpack(colors.neutral)
			elseif UnitIsPVP(${unit}) then
				-- on my team
				r, g, b = unpack(colors.friendly)
			else
				-- either enemy or friend, no violence
				r, g, b = unpack(colors.civilian)
			end
		elseif (UnitIsTapped(${unit}) and not UnitIsTappedByPlayer(${unit})) or UnitIsDead(${unit}) then
			r, g, b = unpack(colors.tapped)
		else
			local reaction = UnitReaction(${unit}, "player")
			if reaction then
				if reaction >= 5 then
					r, g, b = unpack(colors.friendly)
				elseif reaction == 4 then
					r, g, b = unpack(colors.neutral)
				else
					r, g, b = unpack(colors.hostile)
				end
			else
				r, g, b = unpack(colors.unknown)
			end
		end
		
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_FACTION", globals = "UnitIsPlayer;UnitPlayerControlled;UnitCanAttack;UnitIsPVP;UnitIsTapped;UnitIsTappedByPlayer;UnitIsDead;UnitReaction" }

Modifiers.ClassColor = {function(data)
	return [[do
		local _,class = UnitClass(${unit})
		local r, g, b = unpack(colors[class] or colors.unknown)
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", globals = "UnitClass"}

Modifiers.HPColor = {function(data)
	return [[do
		local perc = UnitHealth(${unit}) / UnitHealthMax(${unit})
		local r1, g1, b1
		local r2, g2, b2
		if perc <= 0.5 then
			perc = perc * 2
			r1, g1, b1 = unpack(colors.minHP)
			r2, g2, b2 = unpack(colors.midHP)
		else
			perc = perc * 2 - 1
			r1, g1, b1 = unpack(colors.midHP)
			r2, g2, b2 = unpack(colors.maxHP)
		end
		local r, g, b = r1 + (r2-r1)*perc, g1 + (g2-g1)*perc, b1 + (b2-b1)*perc
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;unpack" }

Modifiers.AggroColor = {function(data)
	return [[do
		local val = UnitReaction("player", ${unit}) or 5

		local info = UnitReactionColor[val]
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_FACTION", globals = "UnitReaction;UnitReactionColor" }

Modifiers.DifficultyColor = {function(data)
	return [[if UnitCanAttack("player", ${unit}) then
		local level = UnitLevel(${unit})
		if level == 0 then
			level = 99
		end
		local info = GetDifficultyColor(level)
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_LEVEL;PLAYER_LEVEL_UP", globals = "UnitCanAttack;UnitLevel;GetDifficultyColor" }

Modifiers.PowerColor = {function(data)
	return [[do
		local powerType = UnitPowerType(${unit})
		local r,g,b
		if powerType == 0 then
			r,g,b = unpack(colors.mana)
		elseif powerType == 1 then
			r,g,b = unpack(colors.rage)
		elseif powerType == 2 then
			r,g,b = unpack(colors.focus)
		elseif powerType == 3 then
			r,g,b = unpack(colors.energy)
		else
			r,g,b = unpack(colors.unknown)
		end
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_DISPLAYPOWER", globals = "unpack;UnitPowerType" }

if date("%Y%m%d") < "20070518" then
	Modifiers.ManaColor = "PowerColor"
end

Modifiers.CombatColor = {function(data)
	return [[
		if UnitAffectingCombat(${unit}) then
			local r,g,b = unpack(colors.inCombat)
	]] .. (data.isMod and [[
			value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
			value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
		else	
	]] .. (data.isMod and [[
			value = "|r" .. value
		]] or [[
			value = "|r"
	]]) .. [[
		end
	]]
end, value = "number;string", ret = "string", events = "Update", globals = "unpack;UnitAffectingCombat" }

Modifiers.Round = {function(data)
	if type(data.arg) ~= "string" then
		local arg = math.floor((data.arg or 0) + 0.5)
		if arg == 0 then
			return [[value = math_floor(value + 0.5)]]
		else
			return [[value = math_floor(value * (10^]] .. arg .. [[) + 0.5) / (10^]] .. arg .. [[)]]
		end
	else -- dynamic
		return [[value = math_floor(value * (10^${arg}) + 0.5) / (10^${arg})]]
	end
end, arg = "nil;number", value = "number", ret = "number", globals = "math.floor"}

Modifiers.Color = {function(data)
	local arg = data.arg:lower()
	if arg:find("%[") and arg:find("%]") then -- dynamic
		return [[do
			local arg = ${arg}
			if not arg:match("^%x%x%x%x%x%x$") then
				arg = "ffffff"
			end
			value = "|cff" .. arg]] .. (isMod and [[ .. value .. "|r"]] or "") .. [[
		end]]
	else
		if not arg:match("^%x%x%x%x%x%x$") then
			arg = "ffffff"
		end
		if data.isMod then
			return [[value = "|cff]] .. arg .. [[" .. value .. "|r"]]
		else
			return [[value = "|cff]] .. arg .. [["]]
		end
	end
end, arg = "string", value = "number;string", ret = "string"}
-- colors
Modifiers.White = "Color(ffffff)"
Modifiers.Red = "Color(ff0000)"
Modifiers.Green = "Color(00ff00)"
Modifiers.Blue = "Color(0000ff)"
Modifiers.Cyan = "Color(00ffff)"
Modifiers.Fuchsia = "Color(ff00ff)"
Modifiers.Yellow = "Color(ffff00)"
Modifiers.Gray = "Color(afafaf)"

local Tags = {}
-- When writing tags, `unit` is modified before the call, that contains the unit id.
-- Also, nil (which translates to empty string), a number, or a string can be returned.
-- the tag itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
-- Note: can take an argument, either nil, a number, or a string.

Tags.IsFriend = {("value = UnitIsFriend('player', ${unit}) and %q"):format(L["True"]), ret = "string;nil", events = "UNIT_FACTION", globals = "UnitIsFriend" }

Tags["~IsFriend"] = {("value = not UnitIsFriend('player', ${unit}) and %q"):format(L["True"]), ret = "string;nil", events = "UNIT_FACTION", globals = "UnitIsFriend" }

Tags.IsEnemy = "~IsFriend"

Tags.Text = {"value = ${arg}", arg = "string", ret = "string"}

Tags.Name = {"value = UnitName(${unit})", ret = "string", events = "UNIT_NAME_UPDATE", globals = "UnitName"}

Tags.NameHostile = "Name:HostileColor"

Tags.Status = {([[do
	if not UnitIsConnected(${unit}) then
		value = %q
	elseif UnitIsDead(${unit}) then
		if UnitIsFriend("player", ${unit}) then
			if DogTag___currentAuras[${unit}]["Interface\\Icons\\Ability_Rogue_FeignDeath"] then
				value = %q
			elseif DogTag___currentAuras[${unit}]["Interface\\Icons\\Spell_Shadow_SoulGem"] then
				value = %q
			end
		end
		if not value then
			value = %q
		end
	elseif UnitIsGhost(${unit}) then
		value = %q
	end
end]]):format(L["Offline"], L["Feigned Death"], L["Soulstone"], L["Dead"], L["Ghost"]), ret = "string;nil", events = "PLAYER_ALIVE;PLAYER_DEAD;PLAYER_UNGHOST", globals = "UnitIsConnected;UnitIsDead;UnitIsGhost;DogTag.__currentAuras" }

Tags["~Status"] = {([[value = UnitIsConnected(${unit}) and not UnitIsDeadOrGhost(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "PLAYER_ALIVE;PLAYER_DEAD;PLAYER_UNGHOST", globals = "UnitIsConnected;UnitIsDeadOrGhost" }

Tags.StatusGone = "Status"

Tags.CurHP = {function(data)
	if MobHealth3 then
		return [[do
			local currValue = UnitHealth(${unit})
			if currValue == 0 then
				value = 0
			else
				if not UnitIsFriend("player", ${unit}) then
					local maxValue = UnitHealthMax(${unit})
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, currValue, maxValue)
					if MHfound then
						value = curr
					end
				end
				if not value then
					value = currValue
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			local currValue = UnitHealth(${unit})
			if currValue == 0 then
				value = 0
			else
				if not UnitIsFriend("player", ${unit}) then
					local name = UnitName(${unit})
					local level = UnitLevel(${unit})
					local ppp = MobHealth_PPP(name..":"..level)
					if ppp > 0 then
						value = math_floor(currValue * ppp + 0.5)
					end
				end
				if not value then
					value = currValue
				end
			end
		end]]
	else
		return [[value = UnitHealth(${unit})]]
	end
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitIsFriend;UnitHealthMax;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.MaxHP = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = 0
			else
				if not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, 1, maxValue)
					if MHfound then
						value = max
					end
				end
				if not value then
					value = maxValue
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			if not UnitIsFriend("player", ${unit}) then
				local name = UnitName(${unit})
				local level = UnitLevel(${unit})
				local ppp = MobHealth_PPP(name..":"..level)
				if ppp > 0 then
					value = math_floor(100 * ppp + 0.5)
				end
			end
			if not value then
				value = UnitHealthMax(${unit})
			end
		end]]
	else
		return [[value = UnitHealthMax(${unit})]]
	end
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitIsFriend;UnitHealthMax;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.PercentHP = {[[do
	local hpmax = UnitHealthMax(${unit})
	if hpmax == 0 then
		value = 0
	else
		value = math_floor(1000 * UnitHealth(${unit}) / hpmax + 0.5) / 10
	end
end]], ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;math.floor" }

Tags.MissingHP = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = 0
			else
				local currValue = UnitHealth(${unit})
				if MobHealth3 and not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, currValue, maxValue)
					if MHfound then
						value = max - curr
					end
				end
				if not value then
					value = maxValue - currValue
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			local currValue = UnitHealth(${unit})
			if currValue == 0 then
				value = 0
			else
				if not UnitIsFriend("player", ${unit}) then
					local name = UnitName(${unit})
					local level = UnitLevel(${unit})
					local ppp = MobHealth_PPP(name..":"..level)
					if ppp > 0 then
						value = math_floor((100 - currValue) * ppp + 0.5)
					end
				end
				if not value then
					value = UnitHealthMax(${unit}) - currValue
				end
			end
		end]]
	else
		return [[value = UnitHealthMax(${unit}) - UnitHealth(${unit})]]
	end
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;UnitIsFriend;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.FractionalHP = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = "0/0"
			else
				local currValue = UnitHealth(${unit})
				if maxValue ~= 100 then
					value = currValue.."/"..maxValue
				elseif MobHealth3 and not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, currValue, maxValue)
					if MHfound then
						value = curr.."/"..max
					end
				end
				if not value then
					value = currValue.."/"..maxValue
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			local currValue = UnitHealth(${unit})
			if not UnitIsFriend("player", ${unit}) then
				local name = UnitName(${unit})
				local level = UnitLevel(${unit})
				local ppp = MobHealth_PPP(name..":"..level)
				if ppp > 0 then
					value = math_floor(currValue * ppp + 0.5) .. "/" .. math_floor(100 * ppp + 0.5)
				end
			end
			if not value then
				value = currValue .. "/" .. UnitHealthMax(${unit})
			end
		end]]
	else
		return [[value = UnitHealth(${unit}).."/"..UnitHealthMax(${unit})]]
	end
end, ret = "string", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;UnitIsFriend;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.SureHP = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue ~= 0 then
				if maxValue ~= 100 then
					value = UnitHealth(${unit}).."/"..maxValue
				elseif MobHealth3 and not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, UnitHealth(${unit}), maxValue)
					if MHfound then
						value = curr.."/"..max
					end
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			if not UnitIsFriend("player", ${unit}) then
				local name = UnitName(${unit})
				local level = UnitLevel(${unit})
				local ppp = MobHealth_PPP(name..":"..level)
				if ppp > 0 then
					value = math_floor(UnitHealth(${unit}) * ppp + 0.5) .. "/" .. math_floor(100 * ppp + 0.5)
				end
			else
				local maxValue = UnitHealthMax(${unit})
				if maxValue ~= 100 then
					value = UnitHealth(${unit}).."/"..maxValue
				end
			end
		end]]
	else
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue ~= 0 and maxValue ~= 100 then
				value = UnitHealth(${unit}).."/"..maxValue
			end
		end]]
	end
end, ret = "string;nil", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;UnitIsFriend;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags["~SureHP"] = {function(data)
	if MobHealth3 then
		return ([[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = %q
			elseif not UnitIsFriend("player", ${unit}) and maxValue == 100 then
				local _, _, MHfound = MobHealth3:GetUnitHealth(${unit}, 1, 100)
				if not MHfound then
					value = %q
				end
			end
		end]]):format(L["True"], L["True"])
	elseif MobHealth_PPP then
		return ([[do
			if not UnitIsFriend("player", ${unit}) then
				local name = UnitName(${unit})
				local level = UnitLevel(${unit})
				local ppp = MobHealth_PPP(name..":"..level)
				if ppp <= 0 then
					value = %q
				end
			elseif UnitHealthMax(${unit}) == 100 then
				value = %q
			end
		end]]):format(L["True"], L["True"])
	else
		return ([[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 or maxValue == 100 then
				value = %q
			end
		end]]):format(L["True"], L["True"])
	end
end, ret = "string;nil", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;UnitIsFriend;MobHealth3;MobHealth_PPP;UnitName;UnitLevel" }

Tags.SmartHP = "SureHP:PercentHP:Percent"

Tags.CurMP = {"value = UnitMana(${unit})", ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana" }

if date("%Y%m%d") < "20070525" then
	Tags.CurMana = "CurMP"
end

Tags.MaxMP = {"value = UnitManaMax(${unit})", ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitManaMax" }

if date("%Y%m%d") < "20070525" then
	Tags.MaxMana = "MaxMP"
end

Tags.PercentMP = {[[do
	local mpmax = UnitManaMax(${unit})
	if mpmax == 0 then
		value = 0
	else
		value = math_floor(1000 * UnitMana(${unit}) / mpmax + 0.5) / 10
	end
end]], ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax;math.floor" }

if date("%Y%m%d") < "20070525" then
	Tags.PercentMana = "PercentMP"
end

Tags.MissingMP = {"value = UnitManaMax(${unit}) - UnitMana(${unit})", ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax" }

if date("%Y%m%d") < "20070525" then
	Tags.MissingMana = "MissingMP"
end

Tags.FractionalMP = {"value = UnitMana(${unit})..'/'..UnitManaMax(${unit})", ret = "string", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax" }

if date("%Y%m%d") < "20070525" then
	Tags.FractionalMana = "FractionalMP"
end

Tags.TypePower = {([[do
	local p = UnitPowerType(${unit})
	if p == 1 then
		value = %q
	elseif p == 2 then
		value = %q
	elseif p == 3 then
		value = %q
	else
		value = %q
	end
end]]):format(L["Rage"], L["Focus"], L["Energy"], L["Mana"]), ret = "string", events = "UNIT_DISPLAYPOWER", globals = "UnitPowerType" }

if date("%Y%m%d") < "20070518" then
	Tags.TypeMana = "TypePower"
end

Tags.IsPowerType = {function(data)
	local arg = data.arg:lower()
	if arg == "rage" then
		return ([[value = UnitPowerType(${unit}) == 1 and %q]]):format(L["Rage"])
	elseif arg == "focus" then
		return ([[value = UnitPowerType(${unit}) == 2 and %q]]):format(L["Focus"])
	elseif arg == "energy" then
		return ([[value = UnitPowerType(${unit}) == 3 and %q]]):format(L["Energy"])
	elseif arg == "mana" then
		return ([[value = UnitPowerType(${unit}) == 0 and %q]]):format(L["Mana"])
	elseif arg:find("%[") and arg:find("%]") then
		return ([[
			local arg = (${arg}):lower()
			if arg == "rage" then
				value = UnitPowerType(${unit}) == 1 and %q
			elseif arg == "focus" then
				value = UnitPowerType(${unit}) == 2 and %q
			elseif arg == "energy" then
				value = UnitPowerType(${unit}) == 3 and %q
			elseif arg == "mana" then
				value = UnitPowerType(${unit}) == 0 and %q
			end
		]]):format(L["Rage"], L["Focus"], L["Energy"], L["Mana"])
	else
		return [[value = nil]]
	end
end, arg = "string", ret = "nil;string", events = "UNIT_DISPLAYPOWER", globals = "UnitPowerType"}

Tags["~IsPowerType"] = {function(data)
	local arg = data.arg:lower()
	if arg == "rage" then
		return ([[value = UnitPowerType(${unit}) ~= 1 and %q]]):format(L["True"])
	elseif arg == "focus" then
		return ([[value = UnitPowerType(${unit}) ~= 2 and %q]]):format(L["True"])
	elseif arg == "energy" then
		return ([[value = UnitPowerType(${unit}) ~= 3 and %q]]):format(L["True"])
	elseif arg == "mana" then
		return ([[value = UnitPowerType(${unit}) ~= 0 and %q]]):format(L["True"])
	elseif arg:find("%[") and arg:find("%]") then
		return ([[
			local arg = (${arg}):lower()
			if arg == "rage" then
				value = UnitPowerType(${unit}) ~= 1 and %q
			elseif arg == "focus" then
				value = UnitPowerType(${unit}) ~= 2 and %q
			elseif arg == "energy" then
				value = UnitPowerType(${unit}) ~= 3 and %q
			elseif arg == "mana" then
				value = UnitPowerType(${unit}) ~= 0 and %q
			end
		]]):format(L["True"], L["True"], L["True"], L["True"])
	else
		return ([[value = %q]]):format(L["True"])
	end
end, arg = "string", ret = "nil;string", events = "UNIT_DISPLAYPOWER", globals = "UnitPowerType"}

Tags.IsRage = "IsPowerType(Rage)"
Tags.IsFocus = "IsPowerType(Focus)"
Tags.IsEnergy = "IsPowerType(Energy)"
Tags.IsMana = "IsPowerType(Mana)"

if date("%Y%m%d") < "20070518" then
	Tags.IsPlayerPowerType = "IsPowerType#player"

	Tags.IsPlayerRage = "IsRage#player"
	Tags.IsPlayerFocus = "IsFocus#player"
	Tags.IsPlayerEnergy = "IsEnergy#player"
	Tags.IsPlayerMana = "IsMana#player"
end

Tags.Level = {[[do
	local x = UnitLevel(${unit})
	if x > 0 then
		value = x
	else
		value = "??"
	end
end]], ret = "number;string", events = "UNIT_LEVEL", globals = "UnitLevel" }

Tags.IsMaxLevel = {([[value = UnitLevel(${unit}) >= %d and %q]]):format(MAX_PLAYER_LEVEL, L["True"]), ret = "string;nil", events = "UNIT_LEVEL", globals = "UnitLevel"}

Tags["~IsMaxLevel"] = {([[value = UnitLevel(${unit}) < %d and %q]]):format(MAX_PLAYER_LEVEL, L["True"]), ret = "string;nil", events = "UNIT_LEVEL", globals = "UnitLevel"}

Tags.Class = {([[value = UnitClass(${unit}) or %q]]):format(UNKNOWN), ret = "string", globals = "UnitClass"}

Tags.SmartClass = {[[value = (UnitIsPlayer(${unit}) or (not UnitIsFriend("player", ${unit}) and not UnitPlayerControlled(${unit}))) and UnitClass(${unit})]], ret = "string;nil", globals = "UnitIsPlayer;UnitIsFriend;UnitPlayerControlled;UnitClass"}

Tags["~SmartClass"] = {([[value = not UnitIsPlayer(${unit}) and (UnitIsFriend("player", ${unit}) or UnitPlayerControlled(${unit})) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPlayer;UnitIsFriend;UnitPlayerControlled;UnitClass"}

Tags.PlayerClass = "IsPlayer?Class"

Tags["~PlayerClass"] = "~IsPlayer"
	
Tags.ShortClass = "Class:ShortClass"

Tags.ShortSmartClass = "SmartClass:ShortClass"

Tags.ShortPlayerClass = "PlayerClass:ShortClass"

Tags.Creature = {([[value = UnitCreatureFamily(${unit}) or UnitCreatureType(${unit}) or %q]]):format(UNKNOWN), ret = "string", globals = "UnitCreatureFamily;UnitCreatureType"}

Tags.Combos = {[[value = GetComboPoints()]], ret = "number", events = "PLAYER_COMBO_POINTS", globals = "GetComboPoints" }

Tags.ComboSymbols = {[[value = ("@"):rep(GetComboPoints())]], ret = "string;nil", events = "PLAYER_COMBO_POINTS", globals = "GetComboPoints"}

Tags["~ComboSymbols"] = {([[value = GetComboPoints() == 0 and %q]]):format(L["True"]), ret = "string;nil", events = "PLAYER_COMBO_POINTS", globals = "GetComboPoints"}

Tags.Classification = {([[do
	local c = UnitClassification(${unit})
	if c == "rare" then
		value = %q
	elseif c == "rareelite" then
		value = %q
	elseif c == "elite" then
		value = %q
	elseif c == "worldboss" then
		value = %q
	end
end]]):format(L["Rare"], L["Rare-Elite"], L["Elite"], L["Boss"]), ret = "string;nil", globals = "UnitClassification"}

Tags["~Classification"] = {([[value = UnitClassification(${unit}) == "normal" and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitClassification"}

Tags.ShortClassification = {([[do
	local c = UnitClassification(${unit})
	if c == "rare" then
		value = %q
	elseif c == "rareelite" then
		value = %q
	elseif c == "elite" then
		value = %q
	elseif c == "worldboss" then
		value = %q
	end
end]]):format(L["r"], L["r+"], L["+"], L["b"]), ret = "string;nil", globals = "UnitClassification"}

Tags["~ShortClassification"] = "~Classification"

Tags.Faction = {[[do
	local _, faction = UnitFactionGroup(${unit})
	value = faction
end]], ret = "string", globals = "UnitFactionGroup"}

Tags.Offline = {([[value = not UnitIsConnected(${unit}) and %q]]):format(L["Offline"]), ret = "string;nil", globals = "UnitIsConnected"}

Tags["~Offline"] = {([[value = UnitIsConnected(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsConnected"}

if date("%Y%m%d") < "20070518" then
	Tags.Connect = "Offline"
end

Tags.AFK = {([[value = UnitIsAFK(${unit}) and %q]]):format(L["AFK"]), ret = "string;nil", events = "PLAYER_FLAGS_CHANGED", globals = "UnitIsAFK"}

Tags["~AFK"] = {([[value = not UnitIsAFK(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "PLAYER_FLAGS_CHANGED", globals = "UnitIsAFK"}

Tags.DND = {([[value = UnitIsDND(${unit}) and %q]]):format(L["DND"]), ret = "string;nil", events = "PLAYER_FLAGS_CHANGED", globals = "UnitIsDND"}

Tags["~DND"] = {([[value = not UnitIsDND(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "PLAYER_FLAGS_CHANGED", globals = "UnitIsDND"}

Tags.AFKDND = {([[value = UnitIsAFK(${unit}) and %q or UnitIsDND(${unit}) and %q]]):format(L["AFK"], L["DND"]), ret = "string;nil", events = "PLAYER_FLAGS_CHANGED", globals = "UnitIsAFK;UnitIsDND"}

Tags["~AFKDND"] = {([[value = not UnitIsAFK(${unit}) and not UnitIsDND(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "PLAYER_FLAGS_CHANGED", globals = "UnitIsAFK;UnitIsDND"}

Tags.SmartRace = {([[
	if UnitIsPlayer(${unit}) then
		value = UnitRace(${unit}) or %q
	else
		value = UnitCreatureFamily(${unit}) or UnitCreatureType(${unit}) or %q
	end
]]):format(UNKNOWN, UNKNOWN), ret = "string", globals = "UnitIsPlayer;UnitRace;UnitCreatureFamily;UnitCreatureType"}

Tags.Race = {"value = UnitRace(${unit})", ret = "string", globals = "UnitRace"}

Tags.ShortRace = "Race:ShortRace"

Tags.PvP = {([[value = UnitIsPVP(${unit}) and %q]]):format(L["PvP"]), ret = "string;nil", globals = "UnitIsPVP"}

Tags["~PvP"] = {([[value = not UnitIsPVP(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPVP"}

Tags.Plus = {([[value = UnitIsPlusMob(${unit}) and %q]]):format(L["+"]), ret = "string;nil", globals = "UnitIsPlusMob"}

Tags["~Plus"] = {([[value = not UnitIsPlusMob(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPlusMob"}

Tags.Sex = {([[do
	local sex = UnitSex(${unit})
	if sex == 0 then
		value = %q
	elseif sex == 1 then
		value = %q
	end
end]]):format(L["Male"], L["Female"]), ret = "string;nil", globals = "UnitSex"}

Tags["~Sex"] = {([[do
	local sex = UnitSex(${unit})
	value = sex ~= 0 and sex ~= 1 and %q
end]]):format(L["True"]), ret = "string;nil", globals = "UnitSex"}

Tags.ShortSex = {([[do
	local sex = UnitSex(${unit})
	if sex == 0 then
		value = %q
	elseif sex == 1 then
		value = %q
	end
end]]):format(L["m"], L["f"]), ret = "string;nil", globals = "UnitSex"}

Tags["~ShortSex"] = "~Sex"

Tags.Resting = {([[value = IsResting() and %q]]):format(L["Resting"]), ret = "string;nil", events = "PLAYER_UPDATE_RESTING", globals = "IsResting" }

Tags["~Resting"] = {([[value = not IsResting() and %q]]):format(L["True"]), ret = "string;nil", events = "PLAYER_UPDATE_RESTING", globals = "IsResting" }

Tags.Leader = {([[value = UnitIsPartyLeader(${unit}) and %q]]):format(L["Leader"]), ret = "string;nil", globals = "UnitIsPartyLeader"}

Tags["~Leader"] = {([[value = not UnitIsPartyLeader(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPartyLeader"}

Tags.LeaderShort = "Leader:Trunc(1)"

Tags.HappyNum = {[[value = GetPetHappiness() or 0]], ret = "number", events = "UNIT_HAPPINESS", globals = "GetPetHappiness"}

Tags.HappyText = {[=[value = _G["PET_HAPPINESS" .. (GetPetHappiness() or 0)]]=], ret = "string;nil", events = "UNIT_HAPPINESS", globals = "_G;GetPetHappiness" }

Tags["~HappyText"] = {([[do
	local happy = GetPetHappiness()
	value = happy ~= 1 and happy ~= 2 and happy ~= 3 and %q
end]]):format(L["True"]), ret = "string;nil", events = "UNIT_HAPPINESS", globals = "GetPetHappiness" }

Tags.HappyIcon = {[[do
	local happy = GetPetHappiness()
	if happy == 3 then
		value = ":D"
	elseif happy == 2 then
		value = ":||"
	elseif happy == 1 then
		value = "B("
	end
end]], ret = "string;nil", events = "UNIT_HAPPINESS", globals = "GetPetHappiness" }

Tags["~HappyIcon"] = "~HappyText"

Tags.CurXP = {[[
	if ${unit} == "pet" then
		value = GetPetExperience()
	else
	 	value = UnitXP(${unit})
	end
]], ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXP" }

Tags.MaxXP = {[[
	if ${unit} == "pet" then
		local _,max = GetPetExperience()
		value = max
	else
	 	value = UnitXPMax(${unit})
	end
]], ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXP" }

Tags.FractionalXP = {[[
	local cur, max
	if ${unit} == "pet" then
		cur, max = GetPetExperience()
	else
	 	cur, max = UnitXP(${unit}), UnitXPMax(${unit})
	end	
	value = cur .. '/' .. max
]], ret = "string", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXP;UnitXPMax" }

Tags.PercentXP = {[[
	if ${unit} == "pet" then
		local cur, max = GetPetExperience()
		if max > 0 then
			value = math_floor(1000 * cur/max + 0.5) / 10
		else
			value = 0
		end
	else
	 	local max = UnitXPMax(${unit})
		if max > 0 then
			local cur = UnitXP(${unit})
			value = math_floor(1000 * cur/max + 0.5) / 10
		else
			value = 0
		end
	end
]], ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXP;math.floor" }

Tags.MissingXP = {[[
	if ${unit} == "pet" then
		local cur, max = GetPetExperience()
		value = max - cur
	else
	 	value = UnitXPMax(${unit}) - UnitXP(${unit})
	end
]], ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXPMax;UnitXP" }

Tags.RestXP = {[[value = ${unit} == "player" and GetXPExhaustion() or 0]], ret = "number", "PLAYER_XP_UPDATE", globals = "GetXPExhaustion"}

Tags.PercentRestXP = {[[do
	local max = UnitXPMax(${unit})
	if max > 0 then
		local rest = ${unit} == "player" and GetXPExhaustion() or 0
		value = math_floor(1000 * rest/max + 0.5) / 10
	else
		value = 0
	end
end]], ret = "number", events = "PLAYER_XP_UPDATE", globals = "GetXPExhaustion;UnitXPMax;math.floor" }


Tags.TappedByMe = {[[value = UnitIsTappedByPlayer(${unit}) and "*"]], ret = "string;nil", events = "Update", globals = "UnitIsTappedByPlayer" }

Tags["~TappedByMe"] = {([[value = not UnitIsTappedByPlayer(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "Update", globals = "UnitIsTappedByPlayer" }

Tags.IsTapped = {([[value = UnitIsTapped(${unit}) and not UnitIsTappedByPlayer(${unit}) and %q]]):format(L["Tapped"]), ret = "string;nil", events = "Update", globals = "UnitIsTapped;UnitIsTappedByPlayer" }

Tags["~IsTapped"] = {([[value = (not UnitIsTapped(${unit}) or UnitIsTappedByPlayer(${unit})) and %q]]):format(L["True"]), ret = "string;nil", events = "Update", globals = "UnitIsTapped;UnitIsTappedByPlayer" }

Tags.IsMaxHP = {([[value = UnitHealth(${unit}) == UnitHealthMax(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealthMax;UnitHealth" }

Tags["~IsMaxHP"] = {([[value = UnitHealth(${unit}) ~= UnitHealthMax(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealthMax;UnitHealth" }

Tags.IsMaxMP = {([[value = UnitMana(${unit}) == UnitManaMax(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax" }

Tags["~IsMaxMP"] = {([[value = UnitMana(${unit}) ~= UnitManaMax(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax" }

if date("%Y%m%d") < "20070525" then
	Tags.IsMaxMana = "IsMaxMP"
end

Tags.HasNoMP = {([[value = UnitManaMax(${unit}) == 0 and %q]]):format(L["True"]), ret = "string;nil", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitManaMax" }

Tags["~HasNoMP"] = {([[value = UnitManaMax(${unit}) ~= 0 and %q]]):format(L["True"]), ret = "string;nil", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitManaMax" }

if date("%Y%m%d") < "20070525" then
	Tags.HasNoMana = "HasNoMP"
end

Tags.InCombat = {([[value = UnitAffectingCombat(${unit}) and %q]]):format(L["Combat"]), ret = "string;nil", events = "Update", globals = "UnitAffectingCombat"}

Tags["~InCombat"] = {([[value = not UnitAffectingCombat(${unit}) and %q]]):format(L["True"]), ret = "string;nil", events = "Update", globals = "UnitAffectingCombat"}

Tags.FKey = {[[do
	local fkey
	if ${unit} == "player" then
		fkey = 0
	else
		fkey = (${unit}):match("^party(%d)$")
	end
	if fkey then
		value = "F" .. (fkey + 1)
	end
end]], ret = "string;nil"}

Tags["~FKey"] = {([[value = ${unit} ~= "player" and not (${unit}):match("^party(%%d)$") and %q]]):format(L["True"]), ret = "string;nil"}

Tags.RaidGroup = {[[
	for i = 1, GetNumRaidMembers() do
		local name, rank, subgroup = GetRaidRosterInfo(i)
		if name == UnitName(${unit}) then
			value = subgroup
			break
		end
	end
]], ret = "string;nil", globals = "GetNumRaidMembers;GetRaidRosterInfo;UnitName"}

Tags["~RaidGroup"] = {([[do
	value = %q
	for i = 1, GetNumRaidMembers() do
		local name, rank, subgroup = GetRaidRosterInfo(i)
		if name == UnitName(${unit}) then
			value = nil
			break
		end
	end
end]]):format(L["True"]), ret = "string;nil", globals = "GetNumRaidMembers;GetRaidRosterInfo;UnitName"}

Tags.Guild = {[[value = GetGuildInfo(${unit})]], ret = "string;nil", globals = "GetGuildInfo"}

Tags["~Guild"] = {([[value = not GetGuildInfo(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "GetGuildInfo"}

Tags.GuildRank = {[[do
	local _, rank = GetGuildInfo(${unit})
	value = rank
end]], ret = "string;nil", globals = "GetGuildInfo"}

Tags["~GuildRank"] = {([[do
	local _, rank = GetGuildInfo(${unit})
	value = not rank and %q
end]]):format(L["True"]), ret = "string;nil", globals = "GetGuildInfo"}

Tags.Target = {[[do
	local u = unit .. "target"
	if UnitExists(u) then
		value = UnitName(u)
	end
end]], ret = "string;nil", events = "UNIT_TARGET", global = "UnitExists;UnitName" }

Tags["~Target"] = {([[do
	local u = unit .. "target"
	value = not UnitExists(u) and %q
end]]):format(L["True"]), ret = "string;nil", events = "UNIT_TARGET", global = "UnitExists;UnitName" }

Tags.IsMouseOver = {([[value = DogTag.__isMouseOver and %q]]):format(L["True"]), ret = "string;nil", events = "Mouseover" }

Tags["~IsMouseOver"] = {([[value = not DogTag.__isMouseOver and %q]]):format(L["True"]), ret = "string;nil", events = "Mouseover" }

Tags.CurRep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		value = repvalue - repmin
	end
end]], ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags["~CurRep"] = "~RepName"

Tags.MaxRep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		value = repmax - repmin
	end
end]], ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags["~MaxRep"] = "~RepName"

Tags.FractionalRep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		repvalue = repvalue - repmin
		repmax = repmax - repmin
		value = repvalue .. '/' .. repmax
	end
end]], ret = "string;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags["~FractionalRep"] = "~RepName"

Tags.PercentRep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		repmax = repmax - repmin
		repvalue = repvalue - repmin
		value = math_floor(1000 * repvalue/repmax + 0.5) / 10
	end
end]], ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo;math.floor"}

Tags["~PercentRep"] = "~RepName"

Tags.MissingRep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		value = repmax - repvalue
	end
end]], ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags["~MissingRep"] = "~RepName"

Tags.RepName = {[[value = GetWatchedFactionInfo()]], ret = "number;nil", events = "Reputation", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags["~RepName"] = {([[value = not GetWatchedFactionInfo() and %q]]):format(L["True"]), ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags.DruidForm = {([[do
	local _,c = UnitClass(${unit})
	if c == "DRUID" then
		local power = UnitPowerType(${unit})
		if power == 1 then
			value = %q
		elseif power == 3 then
			value = %q
		end
	
		DogTag___portrait:SetUnit(${unit})
		local model = DogTag___portrait:GetModel()
		if type(model) == "string" then
			model = model:lower()
			if model:find("owlbear") then
				value = %q
			elseif model:find("sealion") then
				value = %q
			elseif model:find("crow") or model:find("flight") then
				value = %q
			elseif model:find("tiger") then
				value = %q
			elseif model:find("ent") then
				value = %q
			end
		end
	end
end]]):format(L["Bear"], L["Cat"], L["Moonkin"], L["Aquatic"], L["Flight"], L["Travel"], L["Tree"]), ret = "string;nil", events = "UNIT_DISPLAYPOWER;UNIT_PORTRAIT_UPDATE", globals = "UnitClass;UnitPowerType;DogTag.__portrait"}

Tags["~DruidForm"] = {([[do
	local _,c = UnitClass(${unit})
	if c ~= "DRUID" then
		value = %q
	else
		local power = UnitPowerType(${unit})
		if power == 0 then
			DogTag___portrait:SetUnit(${unit})
			local model = DogTag___portrait:GetModel()
			if type(model) == "string" then
				model = model:lower()
				if not model:find("owlbear") and not model:find("sealion") and not model:find("crow") and not model:find("flight") and not model:find("tiger") and not model:find("ent") then
					value = %q
				end
			end
		end
	end
end]]):format(L["True"], L["True"]), ret = "string;nil", events = "UNIT_DISPLAYPOWER;UNIT_PORTRAIT_UPDATE", globals = "UnitClass;UnitPowerType;DogTag.__portrait"}

Tags.IsUnit = {function(data)
	local arg = data.arg:lower()
	if not arg:find("%[") and not arg:find("%]") and not IsLegitimateUnit[arg] then
		return ("value = nil")
	end
	local unit = data.unit
	if unit then
		if not IsLegitimateUnit[unit] then
			return ("value = nil")
		end
		if unit == arg then
			return ("value = %q"):format(L["True"])
		end
		if (unit == "player" or unit == "pet" or unit:find("^party%d$") or unit:find("^partypet%d$")) and (arg == "player" or arg == "pet" or arg:find("^party%d$") or arg:find("^partypet%d$")) then
			return ("value = nil")
		elseif unit:find("^raid%d+") and (arg == "pet" or arg:find("^partypet%d$") or arg:find("^raidpet%d+$")) then
			return ("value = nil")
		elseif unit:find("^raidpet%d+") and (arg == "player" or arg:find("^party%d$") or arg:find("^raid%d+$")) then
			return ("value = nil")
		end
		return ("value = UnitIsUnit(${unit}, ${arg}) and %q"):format(L["True"])
	else
		return ("value = ((${unit} == ${arg}) or UnitIsUnit(${unit}, ${arg})) and %q"):format(L["True"])
	end
end, arg = "string", ret = "string;nil", events = "Special_IsUnit"}

Tags["~IsUnit"] = {function(data)
	local arg = data.arg:lower()
	if not arg:find("%[") and not arg:find("%]") and not IsLegitimateUnit[arg] then
		return ("value = %q"):format(L["True"])
	end
	local unit = data.unit
	if unit then
		if not IsLegitimateUnit[unit] then
			return ("value = %q"):format(L["True"])
		end
		if unit == arg then
			return ("value = %q"):format(L["True"])
		end
		if (unit == "player" or unit == "pet" or unit:find("^party%d$") or unit:find("^partypet%d$")) and (arg == "player" or arg == "pet" or arg:find("^party%d$") or arg:find("^partypet%d$")) then
			return ("value = %q"):format(L["True"])
		elseif unit:find("^raid%d+") and (arg == "pet" or arg:find("^partypet%d$") or arg:find("^raidpet%d+$")) then
			return ("value = %q"):format(L["True"])
		elseif unit:find("^raidpet%d+") and (arg == "player" or arg:find("^party%d$") or arg:find("^raid%d+$")) then
			return ("value = %q"):format(L["True"])
		end
	end
	return ("value = not UnitIsUnit(${unit}, ${arg}) and %q"):format(L["True"])
end, arg = "string", ret = "string;nil", events = "Special_IsUnit"}

Tags.IsPlayer = {([[value = UnitIsPlayer(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPlayer"}

Tags["~IsPlayer"] = {([[value = not UnitIsPlayer(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPlayer"}

Tags.IsPet = {([[value = UnitIsPlayerControlled(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPlayerControlled"}

Tags["~IsPet"] = {([[value = not UnitIsPlayerControlled(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPlayerControlled"}

Tags.IsPlayerOrPet = {([[value = (UnitIsPlayer(${unit}) or UnitIsPlayerControlled(${unit})) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPlayer;UnitIsPlayerControlled"}

Tags["~IsPlayerOrPet"] = {([[value = not UnitIsPlayer(${unit}) and not UnitƒIsPlayerControlled(${unit}) and %q]]):format(L["True"]), ret = "string;nil", globals = "UnitIsPlayer;UnitIsPlayerControlled"}

Tags["PvPRankNum"] = {[[value = UnitPVPRank(${unit})]], ret = "number", globals = "UnitPVPRank"}

Tags["PvPRank"] = {[[do
	local rank = UnitPVPRank(${unit})
	if rank >= 1 then
		value = GetPVPRankInfo(rank, ${unit})
	end
end]], ret = "string;nil", globals = "UnitPVPRank;GetPVPRankInfo"}

Tags._ = {[[value = ${arg}]], arg = "string;number;nil", ret = "string;number;nil"}

Tags.RepColor = Modifiers.RepColor
Tags.StatusColor = Modifiers.StatusColor
Tags.HappyColor = Modifiers.HappyColor
Tags.HostileColor = Modifiers.HostileColor
Tags.ClassColor = Modifiers.ClassColor
Tags.HPColor = Modifiers.HPColor
Tags.AggroColor = Modifiers.AggroColor
Tags.DifficultyColor = Modifiers.DifficultyColor
Tags.PowerColor = Modifiers.PowerColor
if date("%Y%m%d") < "20070518" then
	Tags.ManaColor = Modifiers.ManaColor
end
Tags.CombatColor = Modifiers.CombatColor
Tags.Color = Modifiers.Color
-- colors
Tags.White = Modifiers.White
Tags.Red = Modifiers.Red
Tags.Green = Modifiers.Green
Tags.Blue = Modifiers.Blue
Tags.Cyan = Modifiers.Cyan
Tags.Fuchsia = Modifiers.Fuchsia
Tags.Yellow = Modifiers.Yellow
Tags.Gray = Modifiers.Gray

Tags.SmartMissingHP = "MissingHP:HideZero:Negate"

local function sortStringList(s)
	if not s then
		return nil
	end
	local list = new((";"):split(s))
	table.sort(list)
	local q = table.concat(list, ';')
	list = del(list)
	return q
end

local TagEvents = {}
local ModifierEvents = {}
local TagAliases = {}
local ModifierAliases = {}
local TagArguments = {}
local ModifierArguments = {}
local TagReturns = {}
local ModifierReturns = {}
local ModifierValues = {}
local TagGlobals = {}
local ModifierGlobals = {}
do
	for tag, data in pairs(Tags) do
		if type(data) == "table" then
			local arg = sortStringList(data.arg)
			if arg and arg ~= "nil" then
				TagArguments[tag] = arg
				local a,b,c = (";"):split(arg)
				assert(a == "nil" or a == "number" or a == "string")
				assert(not b or b == "nil" or b == "number" or b == "string")
				assert(not c or c == "nil" or c == "number" or c == "string")
			end
			TagReturns[tag] = sortStringList(data.ret)
			if data.ret then
				local a,b,c = (";"):split(data.ret)
				assert(a == "nil" or a == "number" or a == "string")
				assert(not b or b == "nil" or b == "number" or b == "string")
				assert(not c or c == "nil" or c == "number" or c == "string")
			end
			TagEvents[tag] = sortStringList(data.events)
			TagGlobals[tag] = sortStringList(data.globals)
			local func = data[1]
			if Modifiers[tag] ~= Tags[tag] then
				del(Tags[tag])
			end
			Tags[tag] = func
		elseif type(data) == "string" then
			TagAliases[tag] = data
			Tags[tag] = nil
		end
	end
	
	for mod, data in pairs(Modifiers) do
		if type(data) == "table" then
			local arg = sortStringList(data.arg)
			if arg and arg ~= "nil" then
				ModifierArguments[mod] = arg
				local a,b,c = (";"):split(arg)
				assert(a == "nil" or a == "number" or a == "string")
				assert(not b or b == "nil" or b == "number" or b == "string")
				assert(not c or c == "nil" or c == "number" or c == "string")
			end
			ModifierReturns[mod] = sortStringList(data.ret)
			if data.ret then
				local a,b,c = (";"):split(data.ret)
				assert(a == "nil" or a == "number" or a == "string" or a == "same")
				assert(not b or b == "nil" or b == "number" or b == "string" or b == "same")
				assert(not c or c == "nil" or c == "number" or c == "string" or c == "same")
			end
			local value = sortStringList(data.value)
			if value and value ~= "number;string" then
				ModifierValues[mod] = value
				local a,b = (";"):split(value)
				assert(a == "number" or a == "string")
				assert(not b or b == "number" or b == "string")
			end
			ModifierEvents[mod] = sortStringList(data.events)
			ModifierGlobals[mod] = sortStringList(data.globals)
			local func = data[1]
			del(Modifiers[mod])
			Modifiers[mod] = func
		elseif type(data) == "string" then
			ModifierAliases[mod] = data
			Modifiers[mod] = nil
		end
	end
end

local correctCasing = setmetatable({},{__mode='kv',__index=function(self, tag)
	if not tag then
		return nil
	end
	if Tags[tag] or Modifiers[tag] or TagAliases[tag] or ModifierAliases[tag] then
		self[tag] = tag
		return tag
	end
	local tag_lower = tag:lower()
	for k in pairs(Tags) do
		if k:lower() == tag_lower then
			self[tag] = k
			return k
		end
	end
	for k in pairs(Modifiers) do
		if k:lower() == tag_lower then
			self[tag] = k
			return k
		end
	end
	for k in pairs(TagAliases) do
		if k:lower() == tag_lower then
			self[tag] = k
			return k
		end
	end
	for k in pairs(ModifierAliases) do
		if k:lower() == tag_lower then
			self[tag] = k
			return k
		end
	end
	self[tag] = tag
	return tag
end})

local function toliteral(value)
	if type(value) == "number" then
		return ("%s"):format(value)
	else
		return ("%q"):format(value)
	end
end

local function unpackAndDel(t, start, finish)
	if not start then
		start = 1
	elseif not finish then
		finish = #t
	end
	if start > finish then
		t = del(t)
		return
	end
	return t[start], unpackAndDel(t, start + 1, finish)
end

local function splitCode(code)
	if code:sub(1, 1) == "[" then
		local alpha, bravo, charlie = code:match("^(%b[])([:!%?])(.*)$")
		if alpha then
			return "_(" .. alpha .. ")", bravo, splitCode(charlie)
		end
		local alpha = code:match("^(%b[])$")
		if alpha then
			return splitCode(alpha:sub(2, -2))
		end
	end
	local nextSymbol = code:match("^[^:!%?%(]*(.)")
	if nextSymbol == "(" then
		local alpha, bravo, charlie = code:match("^([^%(]*%b())([:!%?])(.*)$")
		if alpha then
			return alpha, bravo, splitCode(charlie)
		end
		local alpha = code:match("^([^%(]*%b())$")
		if alpha then
			return alpha
		end
	end
	local alpha, bravo, charlie = code:match("^([^:!%?]*)([:!%?])(.*)$")
	if not alpha then
		return code
	end
	return alpha, bravo, splitCode(charlie)
end

local function _splitParam(code)
	local negate, tag, param = code:match("^(~*)([a-zA-Z_#]+)(%b())$")
	if param then
		param = param:sub(2, -2)
	end
	if tag then
		local t, unit = tag:match("([a-zA-Z_]+)#([a-zA-Z_]+)")
		if t then
			return t, param, ((negate:len() % 2) == 1 and true or false), unit:lower()
		else
			return tag, param, ((negate:len() % 2) == 1 and true or false), nil
		end
	else
		negate, tag = code:match("^(~*)(.+)$")
		if tag then
			local t, unit = tag:match("^([a-zA-Z_]+)#([a-zA-Z_]+)$")
			if t then
				return t, nil, ((negate:len() % 2) == 1 and true or false), unit:lower()
			else
				return tag, nil, ((negate:len() % 2) == 1 and true or false), nil
			end
		else
			return code, nil, nil, nil
		end
	end
end
local function splitParam(code)
	local code, param, negate, unit = _splitParam(code)
	code = correctCasing[code]
	if type(param) == "string" and param:find("%[") and param:find("%]") then
		param = DogTag:FixCasing(param)
	end
	if negate then
		local neg_code = '~' .. code
		if Tags[neg_code] or Modifiers[neg_code] then
			return neg_code, param, false, unit
		end
	end
	return code, param, negate, unit
end

local function joinParam(tag, param, negate, unit)
	if tag == "_" and param and param:find("^%[.*%]$") then
		local inner = param:sub(2, -2)
		if not inner:find("[%[%]%?!:]") then
			return inner
		end
		return param
	end
	local t = new()
	if negate then
		t[#t+1] = "~"
	end
	t[#t+1] = tag
	if unit then
		t[#t+1] = "#"
		t[#t+1] = unit
	end
	if param then
		t[#t+1] = "("
		t[#t+1] = param
		t[#t+1] = ")"
	end
	local s = table.concat(t)
	t = del(t)
	return s
end

local parse, compileTree
local function _compileTree(code, globals)
	if type(code) == "string" then
		return code, 0
	end
	local ends
	code[1], ends = _compileTree(code[1], globals)
	if not code[1] then
		code = del(code)
		return nil, ends
	end
	local t = new()
	if not code[2] then
		local tag, param, negate, unit = splitParam(code[1])
		
		local func = Tags[tag]
		if func then
			if param then
				if not TagArguments[tag] then
					t = del(t)
					code = del(code)
					return nil, ("%q Unexpected parameter to %q"):format(param, tag)
				end
				if TagArguments[tag]:find("number") then
					param = tonumber(param) or param
					if type(param) ~= "number" and not TagArguments[tag]:find("string") then
						-- must be number
						t = del(t)
						code = del(code)
						return nil, ("%q Bad param to %q. Expected number."):format(param, tag)
					end
				end
			else
				if TagArguments[tag] and not TagArguments[tag]:find("nil") then
					t = del(t)
					code = del(code)
					return nil, ("%q Expected parameter"):format(tag)
				end
			end
			if unit and unit ~= "player" then
				if not IsLegitimateUnit[unit] then
					t = del(t)
					code = del(code)
					return nil, ("%q Unknown unit"):format(unit)
				end
				t[#t+1] = ("if UnitExists(%q) then "):format(unit)
				ends = ends + 1
			end
			if negate and not TagReturns[tag]:find("nil") then
				t[#t+1] = "value = nil;"
			else
				local g = TagGlobals[tag]
				if g then
					g = new((";"):split(g))
					for _,v in ipairs(g) do
						globals[v] = true
					end
					g = del(g)
				end
			
				if type(func) == "function" then
					local data = new()
					data.isMod = isMod
					data.unit = unit
					data.unit_str = unit and ("%q"):format(unit) or "unit"
					data.arg = param
					data.arg_str = param and toliteral(param) or "nil"
					data.arg_string = param and ("%q"):format(param) or '"nil"'
					data.argtype = ("%q"):format(type(param))
					func = func(data)
					data = del(data)
				end
				if param then
					if type(param) == "string" and param:find("%[") and param:find("%]") then
						t[#t+1] = "local v = value;"
						t[#t+1] = "value = nil;"
						
						t[#t+1] = compileTree(parse(splitCode(param:sub(2, -2))), globals)
						local madeT = false
						local st = 1
						t[#t+1] = [[do ]]
						while st <= param:len() do
							local start, fin, left, literal = param:find("^(.-)(%b[])", st)
							if not start then
								local rest = param:sub(st)
								if rest:len() > 0 then
									if not madeT then
										t[#t+1] = ("value = %q"):format(rest)
									else
										t[#t+1] = ("t[#t+1] = %q;"):format(rest)
									end
								end
								break
							end
							if not madeT and (start ~= 1 or fin ~= param:len()) then
								t[#t+1] = [[local t = new();]]
								madeT = true
							end
							local inner = literal:sub(2, -2)
							st = fin+1
							if left:len() > 0 then
								t[#t+1] = ("t[#t+1] = %q;"):format(left)
							end

							local u = new()
							u[#u+1] = "do "

							local result = compileTree(parse(splitCode(inner)), globals)
							u[#u+1] = result
							if madeT then
								u[#u+1] = [[if value then t[#t+1] = value end;]]
							end
							u[#u+1] = [[end;]]
							local s = table.concat(u)
							u = del(u)
							t[#t+1] = s
						end
						if madeT then
							t[#t+1] = [[value = table_concat(t);]]
							t[#t+1] = [[del(t);]]
						end
						t[#t+1] = [[end;]]
						t[#t+1] = "local arg = value;"
						local a, b, c = TagArguments[tag]:split(';')
						if not c then
							if a == "nil" then
								if b == "number" then
									t[#t+1] = "arg = tonumber(arg);"
								else -- string
									t[#t+1] = "if arg then arg = tostring(arg) end;"
								end
							elseif a == "number" then
								if b == "string" then
									t[#t+1] = "arg = arg or '';"
								else -- not b
									t[#t+1] = "arg = tonumber(arg) or 0"
								end
							else -- a == "string"
								t[#t+1] = "if arg then arg = tostring(arg) else arg = '' end;"
							end
						end
						t[#t+1] = "value = v;"
						func = func:gsub("${arg}", "arg")
						local pos = func:find("${arg:string}")
						if pos and func:find("${arg:string}", pos+1) then
							t[#t+1] = "local arg_string = tostring(arg);"
							func = func:gsub("${arg:string}", "arg_string")
						else
							func = func:gsub("${arg:string}", "tostring(arg)")
						end
						local pos = func:find("${argtype}")
						if pos and func:find("${argtype}", pos+1) then
							t[#t+1] = "local argtype = type(arg);"
							func = func:gsub("${argtype}", "argtype")
						else
							func = func:gsub("${argtype}", "type(arg)")
						end
					else
						func = func:gsub("${arg}", toliteral(param))
						func = func:gsub("${arg:string}", ("%q"):format(param))
						func = func:gsub("${argtype}", ("%q"):format(type(param)))
					end
				else
					func = func:gsub("${arg}", "nil")
					func = func:gsub("${arg:string}", '"nil"')
					func = func:gsub("${argtype}", '"nil"')
				end
				if unit then
					func = func:gsub("${unit}", ("%q"):format(unit))
				else
					func = func:gsub("${unit}", "unit")
				end
				t[#t+1] = func
				if func ~= "" then
					t[#t+1] = ";"
				end
				if negate then
					t[#t+1] = ("value = not value and %q;"):format(L["True"])
				end
			end
--		elseif tag == "" then
--			t[#t+1] = "value = "
--			if negate then
--				t[#t+1] = ("%q"):format(L["True"])
--			else
--				t[#t+1] = "nil"
--			end
--			t[#t+1] = ";"
		else	
			t = del(t)
			code = del(code)
			return nil, ("%q Unknown tag"):format(tag)
		end
	elseif code[2] == "?" then
		t[#t+1] = code[1]
		t[#t+1] = ("if value then value = nil;")
		local q, e = _compileTree(code[3], globals)
		if not q then
			t = del(t)
			code = del(code)
			return nil, e
		end
		t[#t+1] = q
		for i = 1, e do
			t[#t+1] = "end;"
		end
		if code[4] == "!" then
			t[#t+1] = ("else ")
			local q, e = _compileTree(code[5], globals)
			if not q then
				t = del(t)
				code = del(code)
				return nil, e
			end
			t[#t+1] = q
			for i = 1, e do
				t[#t+1] = "end;"
			end
		end
		t[#t+1] = ("end;")
	elseif code[2] == ":" then
		t[#t+1] = code[1]
		local mod, param, negate, unit = splitParam(code[3])
		local func = Modifiers[mod] or Tags[mod]
		if func then
			local isMod = func == Modifiers[mod]
			if param then
				local arg = (isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod])
				if not arg then
					t = del(t)
					code = del(code)
					return nil, ("%q Unexpected param to %q"):format(param, mod)
				end
				
				if arg:find("number") then
					param = tonumber(param) or param
					if type(param) ~= "number" and not arg:find("string") then
						-- must be number
						t = del(t)
						code = del(code)
						return nil, ("%q Bad param to %q. Expected number."):format(param, mod)
					end
				end
			else
				local arg = (isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod])
				if arg and not arg:find("nil") then
					t = del(t)
					code = del(code)
					return nil, ("%q Expected parameter"):format(mod)
				end
			end
			if unit and not IsLegitimateUnit[unit] then
				t = del(t)
				code = del(code)
				return nil, ("%q Unknown unit"):format(unit)
			end
			if isMod then
				local val = ModifierValues[mod]
				if not val then
					t[#t+1] = "if value "
					if unit and unit ~= "player" then
						t[#t+1] = "and UnitExists("
						t[#t+1] = ("%q"):format(unit)
						t[#t+1] = ") "
					end
					t[#t+1] = "then "
				else
					t[#t+1] = [[if type(value) == "]]
					t[#t+1] = val
					t[#t+1] = [[" ]]
					if unit and unit ~= "player" then
						t[#t+1] = "and UnitExists("
						t[#t+1] = ("%q"):format(unit)
						t[#t+1] = ") "
					end
					t[#t+1] = "then "
				end
			else
				t[#t+1] = "if not value "
				if unit and unit ~= "player" then
					t[#t+1] = "and UnitExists("
					t[#t+1] = ("%q"):format(unit)
					t[#t+1] = ") "
				end
				t[#t+1] = "then "
				ends = ends + 1
			end
			if not negate or (isMod and ModifierReturns[mod]:find("nil")) or (not isMod and TagReturns[mod]:find("nil")) then
				if negate and isMod then
					t[#t+1] = "local v = value;"
				end
				local g
				if isMod then
					g = ModifierGlobals[mod]
				else
					g = TagGlobals[mod]
				end
				if g then
					g = new((";"):split(g))
					for _,v in ipairs(g) do
						globals[v] = true
					end
					g = del(g)
				end
				if type(func) == "function" then
					local data = new()
					data.isMod = isMod
					data.unit = unit
					data.unit_str = unit and ("%q"):format(unit) or "unit"
					data.arg = param
					data.arg_str = param and toliteral(param) or "nil"
					data.arg_string = param and ("%q"):format(param) or '"nil"'
					data.argtype = ("%q"):format(type(param))
					func = func(data)
					data = del(data)
				end
				if param then
					if type(param) == "string" and param:find("%[") and param:find("%]") then
						t[#t+1] = "local v = value;"
						t[#t+1] = "value = nil;"
						
						t[#t+1] = compileTree(parse(splitCode(param:sub(2, -2))), globals)
						local madeT = false
						local st = 1
						t[#t+1] = [[do ]]
						while st <= param:len() do
							local start, fin, left, literal = param:find("^(.-)(%b[])", st)
							if not start then
								local rest = param:sub(st)
								if rest:len() > 0 then
									if not madeT then
										t[#t+1] = ("value = %q"):format(rest)
									else
										t[#t+1] = ("t[#t+1] = %q;"):format(rest)
									end
								end
								break
							end
							if not madeT and (start ~= 1 or fin ~= param:len()) then
								t[#t+1] = [[local t = new();]]
								madeT = true
							end
							local inner = literal:sub(2, -2)
							st = fin+1
							if left:len() > 0 then
								t[#t+1] = ("t[#t+1] = %q;"):format(left)
							end

							local u = new()
							u[#u+1] = "do "

							local result = compileTree(parse(splitCode(inner)), globals)
							u[#u+1] = result
							if madeT then
								u[#u+1] = [[if value then t[#t+1] = value end;]]
							end
							u[#u+1] = [[end;]]
							local s = table.concat(u)
							u = del(u)
							t[#t+1] = s
						end
						if madeT then
							t[#t+1] = [[value = table_concat(t);]]
							t[#t+1] = [[del(t);]]
						end
						t[#t+1] = [[end;]]
						t[#t+1] = "local arg = value;"
						local a, b, c = ((isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod])):split(';')
						if not c then
							if a == "nil" then
								if b == "number" then
									t[#t+1] = "arg = tonumber(arg);"
								else -- string
									t[#t+1] = "if arg then arg = tostring(arg) end;"
								end
							elseif a == "number" then
								if b == "string" then
									t[#t+1] = "arg = arg or '';"
								else -- not b
									t[#t+1] = "arg = tonumber(arg) or 0"
								end
							else -- a == "string"
								t[#t+1] = "if arg then arg = tostring(arg) else arg = '' end;"
							end
						end
						t[#t+1] = "value = v;"
						func = func:gsub("${arg}", "arg")
						local pos = func:find("${arg:string}")
						if pos and func:find("${arg:string}", pos+1) then
							t[#t+1] = "local arg_string = tostring(arg);"
							func = func:gsub("${arg:string}", "arg_string")
						else
							func = func:gsub("${arg:string}", "tostring(arg)")
						end
						local pos = func:find("${argtype}")
						if pos and func:find("${argtype}", pos+1) then
							t[#t+1] = "local argtype = type(arg);"
							func = func:gsub("${argtype}", "argtype")
						else
							func = func:gsub("${argtype}", "type(arg)")
						end
					else
						func = func:gsub("${arg}", toliteral(param))
						func = func:gsub("${arg:string}", ("%q"):format(param))
						func = func:gsub("${argtype}", ("%q"):format(type(param)))
					end
				else
					func = func:gsub("${arg}", "nil")
					func = func:gsub("${arg:string}", '"nil"')
					func = func:gsub("${argtype}", '"nil"')
				end
				if unit then
					func = func:gsub("${unit}", ("%q"):format(unit))
				else
					func = func:gsub("${unit}", "unit")
				end
				t[#t+1] = func
				if func ~= "" then
					t[#t+1] = ";"
				end
				if negate then
					t[#t+1] = "value = not value and "
					if isMod then
						t[#t+1] = "v"
						t[#t+1] = " end;"
					else
						t[#t+1] = ("%q;"):format(L["True"])
					end
				elseif isMod then
					t[#t+1] = " end;"
				end
			else
				t[#t+1] = "value = nil;"
			end
		else
			t = del(t)
			code = del(code)
			return nil, ("%q Unknown modifier or tag"):format(mod)
		end	
	else
		t = del(t)
		code = del(code)
		return nil, ("%q Unknown separator"):format(code[2])
	end
	code = del(code)
	local s = table.concat(t)
	t = del(t)
	return s, ends
end

function compileTree(code, globals)
	local q, e = _compileTree(code, globals)
	if not q then
		return ("value = %q;"):format(e)
	end
	for i = 1, e do
		q = q .. "end;"
	end
	return q
end

local function select2(min, max, ...)
	if min <= max then
		return select(min, ...), select2(min+1, max, ...)
	end
end

function parse(...)
	local n = select('#', ...)
	if n == 1 then
		return new((...))
	end
	local q_num = 0
	local q_pos
	local x_pos
	for i = 1, n do
		if select(i, ...) == "?" then
			q_num = q_num + 1
			if not q_pos then
				q_pos = i
			end
		elseif select(i, ...) == "!" then
			q_num = q_num - 1
			if q_num == 0 then
				x_pos = i
				break
			end
		end
	end
	if q_pos then
		if not x_pos then
			return new(parse(select2(1, q_pos - 1, ...)), "?", parse(select(q_pos + 1, ...)))
		else
			return new(parse(select2(1, q_pos - 1, ...)), "?", parse(select2(q_pos + 1, x_pos - 1, ...)), "!", parse(select(x_pos + 1, ...)))
		end
	end
	return new(parse(select2(1, n - 2, ...)), ":", select(n, ...))
end

local string_gsub = string.gsub
local function cleanText(text)
	while true do
		local lastText = text
		text = string_gsub(text, " +$", "")
		text = string_gsub(text, "^ +", "")
		text = string_gsub(text, "  +", " ")
		text = string_gsub(text, "|| ||", "")
		text = string_gsub(text, "^||", "")
		text = string_gsub(text, "||$", "")
		if text == lastText then
			return text
		end
	end
end

local UnitToLocale = {player = L["Player"], target = L["Target"], pet = L["%s's pet"]:format(L["Player"]), focus = L["Focus"], mouseover = L["Mouse-over"]}
setmetatable(UnitToLocale, {__index=function(self, unit)
	if unit:find("pet$") then
		local nonPet = unit:sub(1, -4)
		self[unit] = L["%s's pet"]:format(self[nonPet])
		return self[unit]
	elseif not unit:find("target$") then
		if unit:find("^party%d$") then
			local num = unit:match("^party(%d)$")
			self[unit] = L["Party member #%d"]:format(num)
			return self[unit]
		elseif unit:find("^raid%d%d?$") then
			local num = unit:match("^raid(%d%d?)$")
			self[unit] = L["Raid member #%d"]:format(num)
			return self[unit]
		elseif unit:find("^partypet%d$") then
			local num = unit:match("^partypet(%d)$")
			self[unit] = UnitToLocale["party" .. num .. "pet"]
			return self[unit]
		elseif unit:find("^raidpet%d%d?$") then
			local num = unit:match("^raidpet(%d%d?)$")
			self[unit] = UnitToLocale["raid" .. num .. "pet"]
			return self[unit]
		end
		self[unit] = unit
		return unit
	end
	local nonTarget = unit:sub(1, -7)
	self[unit] = L["%s's target"]:format(self[nonTarget])
	return self[unit]
end})

local codeToFunction = setmetatable({}, {__index=function(self, code)
	if cleanText(code) == "" then
		self[code] = function()
			return ""
		end
		return self[code]
	end
	local t = new()
	t[#t+1] = ([[local DogTag = AceLibrary(%q);]]):format(MAJOR_VERSION)
	t[#t+1] = [[local Tags = DogTag.__Tags;local Modifiers = DogTag.__Modifiers;local new = DogTag.__new;local del = DogTag.__del;local colors = DogTag.__colors;]]
	t[#t+1] = "" -- placeholder
	t[#t+1] = [[local cleanText = DogTag.__cleanText;]]
	t[#t+1] = [[return function(unit) ]]
	t[#t+1] = [[if not UnitExists(unit) then return ]]
	local globals = new()
	globals['table.concat'] = true
	globals['string.gsub'] = true
	globals['type'] = true
	globals['tonumber'] = true
	globals['UnitExists'] = true
	if code:find("[%[%?:!][Nn][Aa][Mm][Ee][%]:%?!]") then
		t[#t+1] = [=[DogTag___UnitToLocale[unit]]=]
		globals['DogTag.__UnitToLocale'] = true
	else
		t[#t+1] = [[""]]
	end
	t[#t+1] = [[ end;]]
	local madeTmp = false
	local st = 1
	while st <= code:len() do
		local start, fin, left, literal = code:find("^(.-)(%b[])", st)
		if not start then
			local rest = code:sub(st)
			if rest:len() > 0 then
				if not madeTmp then
					t[#t+1] = ("return %q"):format(rest)
				else
					t[#t+1] = ("tmp[#tmp+1] = %q;"):format(rest)
				end
			end
			break
		end
		if not madeTmp and (start ~= 1 or fin ~= code:len()) then
			t[#t+1] = [[local tmp = new();]]
			madeTmp = true
		end
		local inner = literal:sub(2, -2)
		st = fin+1
		if left:len() > 0 then
			t[#t+1] = ("tmp[#tmp+1] = %q;"):format(left)
		end
		
		local u = new()
		u[#u+1] = "do local value;"
		
		local result = compileTree(parse(splitCode(inner)), globals)
		u[#u+1] = result
		if not madeTmp then
			u[#u+1] = [[return value;]]
		else
			u[#u+1] = [[if value then tmp[#tmp+1] = value end;]]
		end
		u[#u+1] = [[end;]]
		local s = table.concat(u)
		u = del(u)
		t[#t+1] = s
	end
	local u = new()
	for k in pairs(globals) do
		u[#u+1] = "local "
		u[#u+1] = k:gsub("%.", "_")
		u[#u+1] = " = "
		u[#u+1] = k
		u[#u+1] = ";"
	end
	globals = del(globals)
	t[3] = table.concat(u)
	u = del(u)
	if madeTmp then
		t[#t+1] = [[local s = table_concat(tmp);]]
		t[#t+1] = [[del(tmp);]]
		t[#t+1] = [[return cleanText(s);]]
	end
	t[#t+1] = [[end;]]
	local s = table.concat(t)
	t = del(t)
	DogTag.__Tags = Tags
	DogTag.__Modifiers = Modifiers
	DogTag.__new = new
	DogTag.__del = del
	DogTag.__colors = colors
	DogTag.__portrait = portrait
	DogTag.__cleanText = cleanText
	DogTag.__UnitToLocale = UnitToLocale
	DogTag.__currentAuras = currentAuras
	local func = loadstring(s)
	if not func then
		error(("DogTag-1.0: Error with code %q. Please inform ckknight."):format(code))
	else
		val = func()
	end
	DogTag.__Tags = nil
	DogTag.__Modifiers = nil
	DogTag.__new = nil
	DogTag.__del = nil
	DogTag.__colors = nil
	DogTag.__portrait = nil
	DogTag.__cleanText = nil
	DogTag.__UnitToLocale = nil
	DogTag.__currentAuras = nil
	self[code] = val
	return val
end, __mode='k'})

local codeToEventList; codeToEventList = setmetatable({}, {__index = function(self, key)
	local t = new()
	local len = key:len()
	local st = 1
	while st <= len do
		local start, fin, left, literal = key:find("^(.-)(%b[])", st)
		if not start then
			break
		end
		local inner = literal:sub(2, -2)
		st = fin+1
		local code = new(splitCode(inner))
		for i = 1, #code, 2 do
			local bit, param, negate, unit = splitParam(code[i])
			local func
			if code[i-1] ~= ":" then
				func = Tags[bit]
			else
				func = Modifiers[bit] or Tags[bit]
			end
			local isMod = func == Modifiers[bit]
			local eventList = (isMod and ModifierEvents[bit]) or (not isMod and TagEvents[bit])
			if eventList then
				local events = new((";"):split(eventList))
				for _,event in ipairs(events) do
					if event == "Special_IsUnit" then
						if param == "player" or param == "pet" or param == "target" or param == "focus" or param == "mouseover" or param:find("^party%d$") or param:find("^partypet%d$") or param:find("^raid%d+$") or param:find("^raidpet%d+$") then
							t[event] = param
							t[event.."Two"] = unit or true
						else
							t["Update"] = param
						end
					else
						local e, u = event:match("([A-Z_]+)%(([a-z]+)%)")
						if e then
							t[e] = u
						else
							t[event] = unit or true
						end
					end
				end
				events = del(events)
			end
			if type(param) == "string" and param:find("%[") and param:find("%]") then
				for k, v in pairs(codeToEventList[param]) do
					if t[k] then
						if t[k] ~= v then
							t[k .. v] = v
						end
					else
						t[k] = v
					end
				end
			end
		end
		code = del(code)
	end
	self[key] = t
	return t
end, __mode = 'k'})

local antiAlias
do
	local function handler(literal)
		local inner = literal:sub(2,-2)
		local code = new(splitCode(inner))
		local real = new()
		for i = 1, #code, 2 do
			local bit, param, negate, unit = splitParam(code[i])
			local func
			local isMod = false
			if code[i-1] ~= ":" then
				func = TagAliases[bit]
			else
				func = ModifierAliases[bit]
				if not func and not Modifiers[bit] then
					func = TagAliases[bit]
				else
					isMod = true
				end
			end
			if i == 1 then
				real[1] = "["
			else
				real[#real+1] = code[i-1]
			end
			if negate then
				real[#real+1] = "~"
			end
			if type(param) == "string" and param:find("%[") and param:find("%]") then
				param = antiAlias[param]
			end
			if func then
				local group = not isMod and func:find("[%?!]")
				if group then
					real[#real+1] = "["
				end
				if param and func:find("[:%(%?!]") then
					if func:find("%(arg%)") then
						real[#real+1] = func:gsub("%(arg%)", "(" .. param .. ")")
					else
						real[#real+1] = joinParam(bit, param, nil, unit)
					end
				elseif unit and func:find("#") then
					real[#real+1] = joinParam(bit, param, nil, unit)
				elseif not func:find("[:%?!]") then
					local fbit, fparam, fnegate, funit = splitParam(func)
					real[#real+1] = joinParam(fbit, antiAlias[fparam] or param, fnegate, funit or unit)
				elseif unit then
					if param then
						real[#real+1] = joinParam(bit, param, nil, unit)
					else
						local fcode = new(splitCode(func))
						for j = 1, #fcode, 2 do
							if j > 1 then
								real[#real+1] = fcode[j-1]
							end
							local fbit, fparam, fnegate, funit = splitParam(fcode[j])
							real[#real+1] = joinParam(fbit, antiAlias[fparam], fnegate, unit)
						end
					end
				else
					real[#real+1] = func
					if param then
						real[#real+1] = "("
						real[#real+1] = param
						real[#real+1] = ")"
					end
				end
				if group then
					real[#real+1] = "]"
				end
			else	
				real[#real+1] = joinParam(bit, param, nil, unit)
			end
		end
		code = del(code)
		
		real[#real+1] = ']'
		local str = table.concat(real):gsub("~~", "")
		real = del(real)
		return str
	end
	antiAlias = setmetatable({}, {__index = function(self, key)
		if key == nil then
			return nil
		end
		local value = key:gsub("~~", "")
		while true do
			local lastValue = value
			value = lastValue:gsub("(%b[])", handler)
			if value == lastValue then
				break
			end
		end
		self[key] = value
		return value
	end, __mode= 'k'})
end

local eventData = setmetatable({}, {__index = function(self, event)
	local t = new()
	self[event] = t
	if event:find("^[A-Z_]+$") then
		DogTag.frame:RegisterEvent(event)
	end
	return t
end})
local buckets = setmetatable({}, {__index = function(self, event)
	local t = new()
	self[event] = t
	return t
end})

local registry
local codeHash

local frameToText = {}
local textToFrame = {}

local function updateText(text, code, unit)
	local success, ret = pcall(codeToFunction[code], unit)
	if success then
		text:SetText(ret)
	else
		geterrorhandler()(("%s.%d: Error with code %q on %q. %s"):format(MAJOR_VERSION, MINOR_VERSION, code, unit, ret))
	end
end

local NormalUnits = {player = true, pet = true, target = true, mouseover = true, focus = true}
for i = 1, 4 do
	NormalUnits['party' .. i] = true
	NormalUnits['partypet' .. i] = true
end
for i = 1, 40 do
	NormalUnits['raid' .. i] = true
end
function DogTag:RegisterFontString(text, frame, unit, code)
	self:argCheck(text, 2, "table")
	self:argCheck(frame, 3, "table")
	self:argCheck(unit, 4, "string")
	self:argCheck(code, 5, "string")
	if not IsLegitimateUnit[unit] then
		code = ""
	end
	code = antiAlias[code]
	if codeHash[text] then
		if codeHash[text] == code and registry[text] == unit and textToFrame[text] == frame then
			if text:IsShown() and textToFrame[text]:IsShown() then
				DogTag.__isMouseOver = frame == lastMouseFocus
--				text:SetText(codeToFunction[code](unit))
				updateText(text, code, unit)
			end
			return
		end
		self:UnregisterFontString(text)
	end
	codeHash[text] = code
	registry[text] = unit
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event, u in pairs(events) do
			if u == true then
				eventData[event][text] = unit
			else
				eventData[event][text] = u
			end
		end
	end
	registry[text] = unit
	
	if not frameToText[frame] then
		frameToText[frame] = new()
	end
	frameToText[frame][text] = unit
	textToFrame[text] = frame
	
	if text:IsShown() and textToFrame[text]:IsShown() then
		DogTag.__isMouseOver = frame == lastMouseFocus
--		text:SetText(codeToFunction[code](unit))
		updateText(text, code, unit)
	end
end

function DogTag:UnregisterFontString(text)
	self:argCheck(text, 2, "table")
	local code = codeHash[text]
	if not code then
		return
	end
	
	local unit = registry[text]
	registry[text] = nil
	codeHash[text] = nil
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			local t = eventData[event]
			assert(t)
			t[text] = nil
			if not next(t) then
				eventData[event] = del(t)
				if event:find("^[A-Z_]+$") then
					self.frame:UnregisterEvent(event)
					buckets[event] = del(buckets[event])
				end
			end
		end
	end
	
	text:SetText("")
	
	local frame = textToFrame[text]
	textToFrame[text] = nil
	frameToText[frame][text] = nil
	if not next(frameToText[frame]) then
		frameToText[frame] = del(frameToText[frame])
	end
end

function DogTag:UpdateFontString(text)
	self:argCheck(text, 2, "table")
	local unit = registry[text]
	local frame = textToFrame[text]
	if unit and text:IsShown() and frame:IsShown() then
		DogTag.__isMouseOver = frame == lastMouseFocus
--		text:SetText(codeToFunction[codeHash[text]](unit))
		updateText(text, codeHash[text], unit)
	end
end

function DogTag:UpdateAllForUnit(unit)
	self:argCheck(unit, 2, "string")
	if not IsLegitimateUnit[unit] then
		return
	end
	
	local toUpdate = new()
	for text, u in pairs(registry) do
		if u == unit then
			toUpdate[text] = unit
		end
	end
	for event, data in pairs(eventData) do
		for text, u in pairs(data) do
			if u == unit then
				toUpdate[text] = registry[text]
			end
		end
	end
	for text, u in pairs(toUpdate) do
		local frame = textToFrame[text]
		if text:IsShown() and frame:IsShown() then
			DogTag.__isMouseOver = frame == lastMouseFocus
--			text:SetText(codeToFunction[codeHash[text]](u))
			updateText(text, codeHash[text], u)
		end
	end
	toUpdate = del(toUpdate)
end

function DogTag:UpdateAllForFrame(frame)
	self:argCheck(frame, 2, "table")
	
	if not frame:IsShown() then
		return
	end
	
	local list = frameToText[frame]
	if not list then
		return
	end
	
	for text, unit in pairs(list) do
		if text:IsShown() then
			DogTag.__isMouseOver = frame == lastMouseFocus
--			text:SetText(codeToFunction[codeHash[text]](unit))
			updateText(text, codeHash[text], unit)
		end
	end
end

function DogTag:Evaluate(unit, code)
	if not IsLegitimateUnit[unit] then
		return ""
	end
	DogTag.__isMouseOver = false
	return codeToFunction[antiAlias[code]](unit)
end

do
	local function handler(literal)
		local inner = literal:sub(2, -2)
		local code = new(splitCode(inner))
		local real = new()
		for i = 1, #code, 2 do
			local bit, param, negate, unit = splitParam(code[i])
			
			if i == 1 then
				real[1] = "["
			else
				real[#real+1] = code[i-1]
			end
			
			real[#real+1] = joinParam(bit, param, negate, unit)
		end
		real[#real+1] = ']'
		local str = table.concat(real):gsub("~~", "")
		real = del(real)
		return str
	end
	function DogTag:FixCasing(code)
		return (code:gsub("(%b[])", handler))
	end
end

function DogTag:PARTY_MEMBERS_CHANGED()
	for text, unit in pairs(registry) do
		local frame = textToFrame[text]
		if text:IsShown() and frame:IsShown() then
			DogTag.__isMouseOver = frame == lastMouseFocus
--			text:SetText(codeToFunction[codeHash[text]](unit))
			updateText(text, codeHash[text], unit)
		end
	end
end
DogTag.PARTY_LEADER_CHANGED = DogTag.PARTY_MEMBERS_CHANGED
DogTag.RAID_ROSTER_UPDATE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_MEMBER_ENABLE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_MEMBER_DISABLE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_ENTERING_WORLD = DogTag.PARTY_MEMBERS_CHANGED
DogTag.UNIT_PET = DogTag.PARTY_MEMBERS_CHANGED

function DogTag:PLAYER_TARGET_CHANGED()
	self:UpdateAllForUnit('target')
end

function DogTag:PLAYER_FOCUS_CHANGED()
	self:UpdateAllForUnit('focus')
end

function DogTag:PLAYER_PET_CHANGED()
	self:UpdateAllForUnit('pet')
end

local inMouseover = false
function DogTag:UPDATE_MOUSEOVER_UNIT()
	inMouseover = true
	self:UpdateAllForUnit('mouseover')
end

function DogTag:PLAYER_LOGIN()
	-- for extra addons' help, such as MobHealth3 or MobInfo2
	for k in pairs(codeToFunction) do
		codeToFunction[k] = nil
	end
	self:PARTY_MEMBERS_CHANGED()
end

function DogTag:SetColorConstantTable(t)
	colors = t
	-- need to update the colors reference in the functions
	for k in pairs(codeToFunction) do
		codeToFunction[k] = nil
	end
	self:PARTY_MEMBERS_CHANGED()
end

function DogTag:OnMouseoverUpdate()
	local toUpdate = new()
	local mouseFocus = GetMouseFocus()
	local list, list2 = frameToText[lastMouseFocus], frameToText[mouseFocus]
	lastMouseFocus = mouseFocus
	if list then
		if list2 then
			for text, unit in pairs(eventData.Mouseover) do
				if list[text] or list2[text] then
					toUpdate[text] = unit
				end
			end
		else
			for text, unit in pairs(eventData.Mouseover) do
				if list[text] then
					toUpdate[text] = unit
				end
			end
		end
	elseif list2 then
		for text, unit in pairs(eventData.Mouseover) do
			if list2[text] then
				toUpdate[text] = unit
			end
		end
	end
	
	for text, u in pairs(registry) do
		if u == "mouseover" then
			toUpdate[text] = "mouseover"
		end
	end	
	for event, data in pairs(eventData) do
		for text, u in pairs(data) do
			if u == "mouseover" then
				toUpdate[text] = registry[text]
			end
		end
	end
	
	for text, u in pairs(toUpdate) do
		local frame = textToFrame[text]
		if text:IsShown() and frame:IsShown() then
			DogTag.__isMouseOver = frame == lastMouseFocus
--			text:SetText(codeToFunction[codeHash[text]](u))
			updateText(text, codeHash[text], u)
		end
	end
	toUpdate = del(toUpdate)
end

function DogTag:OnReputationUpdate()
	for text, unit in pairs(eventData.Reputation) do
		if text:IsShown() and textToFrame[text]:IsShown() then
--			text:SetText(codeToFunction[codeHash[text]](unit))
			updateText(text, codeHash[text], unit)
		end
	end
end

local auraQueue = {}
function DogTag:UNIT_AURA(unit)
	auraQueue[unit] = true
end

local function activate(self, oldLib, oldDeactivate)
	DogTag = self
	
	self.frame = oldLib and oldLib.frame or CreateFrame("Frame", "DogTag10Frame")
	
	self.frame:UnregisterAllEvents()
	self.frame:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self.frame:RegisterEvent("PARTY_LEADER_CHANGED")
	self.frame:RegisterEvent("RAID_ROSTER_UPDATE")
	self.frame:RegisterEvent("PARTY_MEMBER_ENABLE")
	self.frame:RegisterEvent("PARTY_MEMBER_DISABLE")
	self.frame:RegisterEvent("PARTY_ENTERING_WORLD")
	self.frame:RegisterEvent("PLAYER_TARGET_CHANGED")
	self.frame:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self.frame:RegisterEvent("PLAYER_PET_CHANGED")
	self.frame:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	self.frame:RegisterEvent("UNIT_PET")
	self.frame:RegisterEvent("PLAYER_LOGIN")
	self.frame:RegisterEvent("UNIT_AURA")
	self.frame:SetScript("OnEvent", function(this, event, arg1, ...)
		local func = self[event]
		if func then
			func(self, arg1, ...)
			return
		end	
		buckets[event][arg1 or false] = true
	end)
	local timePassed = 0
	local num = 0
	local toUpdate = {}
	self.frame:SetScript("OnUpdate", function(this, elapsed)
		timePassed = timePassed + elapsed
		if timePassed > 0.05 then
			timePassed = 0
			if inMouseover and not UnitExists("mouseover") then
				inMouseover = false
				for text, u in pairs(registry) do
					if u == "mouseover" then
						toUpdate[text] = true
					end
				end
				for event, data in pairs(eventData) do
					for text, u in pairs(data) do
						if u == "mouseover" then
							toUpdate[text] = true
						end
					end
				end
			end
			num = num + 1
			if num%5 == 0 then
				if num%20 == 0 then
					for unit in pairs(currentAuras) do
						if not NormalUnits[unit] then
							currentAuras[unit] = del(currentAuras[unit])
						end
					end
				end
				for unit in pairs(auraQueue) do
					auraQueue[unit] = nil
					local t = new()
					for i = 1, 40 do
						local _,_,texture = UnitBuff(unit, i)
						if not texture then
							break
						end
						t[texture] = true
					end
					for i = 1, 40 do
						local _,_,texture = UnitDebuff(unit, i)
						if not texture then
							break
						end
						t[texture] = true
					end
					local old = rawget(currentAuras, unit) or new()
					local changed = new()
					for k in pairs(t) do
						if not old[k] then
							changed[k] = true
						end
					end
					for k in pairs(old) do
						if not t[k] then
							changed[k] = true
						end
					end
					currentAuras[unit] = t
					old = del(old)
					for texture in pairs(changed) do
						if eventData[texture] then
							for text in pairs(eventData[texture]) do
								toUpdate[text] = true
							end
						end
					end
				end
			end
			if num%3 == 0 then
				if eventData.Update then
					for text in pairs(eventData.Update) do
						toUpdate[text] = true
					end
				end
			end
			for event, units in pairs(buckets) do
				if units[false] then
					for text in pairs(eventData[event]) do
						toUpdate[text] = true
					end
				else
					for text, unit in pairs(eventData[event]) do
						if units[unit] then
							toUpdate[text] = true
						end
					end
				end
				buckets[event] = del(units)
			end
			for text in pairs(toUpdate) do
				toUpdate[text] = nil
				local unit = registry[text]
				local frame = textToFrame[text]
				if unit and text:IsShown() and frame:IsShown() then
					DogTag.__isMouseOver = frame == lastMouseFocus
					updateText(text, codeHash[text], unit)
				end
			end
		end
	end)
	
	if not oldLib or not oldLib.hookedRep then
		hooksecurefunc("ReputationWatchBar_Update", function()
			self:OnReputationUpdate()
		end)
	end
	self.hookedRep = true
	
	registry = {}
	self.registry = registry
	codeHash = {}
	self.codeHash = codeHash
	if oldLib then
		for text, unit in pairs(oldLib.registry) do
			self:RegisterFontString(text, unit, oldLib.codeHash[text])
		end
	end
end

AceLibrary:Register(DogTag, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
